# Install
`npm install git+ssh://git@gitlab.bigstep.com:17734/npm/encryption.git#1.0.0`

# Encryption

Encryption utilities for NodeJS. Uses [node-forge](https://www.npmjs.com/package/node-forge).

## Format

`KEY_INDEX_NAME|ALGORITHM_NAME_LOWER_CASE|base64(IV + DATA_HASH + cipher)`

If `keys_partition` property is set as an non-empty `string`, than it will be prepended to `KEY_INDEX_NAME`.

The IV (AES initialization vector) is an 16 bytes random Buffer.
The hash is an MD5 hash of [original data concatenated with a salt]. It is used to check for decryption errors.

Encryption output example: `"k0|aes-cbc|someBase64Cipher"` or `"partition-1_k0|aes-cbc|someBase64Cipher"`

## Config object

```js
{
	active_index: "KEY_INDEX_NAME_1",
	keys_partition: "KEY_PARTITION" | null | undefined, //if set with a non-empty "string", it will be added as a prefix to all keys (except: "default"). 
	keys: {
		"KEY_INDEX_NAME_1": {
			//https://www.npmjs.com/package/node-forge#cipher
			// Note: a key size of 16 bytes will use AES-128, 24 => AES-192, 32 => AES-256
			["ALGORITHM_NAME_LOWER_CASE_WITHOUT_MODE" + "_key"]: "KEY_VALUE", //Example: "aes_key"
			salt: "SALT_VALUE",
			created: "YYYY-MM-DDTHH:MM:SSZ",
		},
		"KEY_INDEX_NAME_2": {
			["ALGORITHM_NAME_LOWER_CASE_WITHOUT_MODE" + "_key"]: "KEY_VALUE",
			salt: "SALT_VALUE",
			created: "YYYY-MM-DDTHH:MM:SSZ",
			compromised: true | false //Optional - An Error is thrown when trying to encrypt or decrypt anything with this key.
		},
		...
		//Optional - The "default" key is used for decryption when there is no prefix set on the encrypted data with the key index name.
		"default": {
			["ALGORITHM_NAME_LOWER_CASE_WITHOUT_MODE" + "_key"]: "KEY_VALUE",
			salt: "SALT_VALUE",
			created: "YYYY-MM-DDTHH:MM:SSZ"
		}
	}
}
```

When using an AES algorithm, the key size will determine whether AES-128, AES-192, or AES-256 is used (all are supported).

### Notes

**Active or default keys cannot be set as _compromised_.** Create a new encryption key and set it as active or default and use rotation for updating all compromised values.

## Decryption options

Option name						| Type					| Default value | Description
--- 							| ---					|--- 			| ---
skipHashValidation 				|	`boolean`			| `false` 		| Specifies if the decrypted data hash should be validated against the hash from the original encrypted data. It can be set to `true` if the encrypted data does not contain a hash in payload.
allowCompromisedKeys 			|	`boolean`			| `false` 		| Allows decryption to use compromised keys. Default, it throws and error when trying to decrypt data using a compromised key.
payloadBase64Encoded 			|	`boolean`			| `true` 		| Specifies if the payload is `base64` encoded or it is a raw `binary` data.
hasPayloadHash 					|	`boolean`			| `true` 		| Specifies if the hash is in the payload. If set to `false`, then `skipHashValidation` is set to `true` and hash is ignored.
hasPayloadIV 					|	`boolean`			| `true` 		| Specifies if the IV is in the payload. If set to `false`, then `customIV` must be set as a **non-empty** `string` or `Buffer`. That IV will be used for decryption.
customIV 						|	`string` or `Buffer`| `null`		| It is used only when `hasPayloadIV` is set to `false`.
offsetBytesBeforePayloadParse 	|	`Integer`			| `0` 			| If set to a positive `Integer`, then that number of bytes will be skipped from payload before parsing it *(right AFTER base64 decode)*.
parseMetadata 					|	`boolean`			| `true`		| Specifies if decrypt should try to parse encrypted data metadata.

## Usage example

```js
const strTestFeedData = "something";

let objEncryptionKeysConfig = {
	active_index: "k1",
	keys_partition: "partition-1",
	keys: {
		"k0": {
			aes_key: "H9Tgfjmo91Jaj12kjbkPE89Rj89fasjq",
			salt: "SOMErondomSTUFF",
			created: "2018-12-31T23:51:55Z"
		},
		"k1": {
			aes_key: "RXTgfjmo91Jaj12kjbkPE89Rj89fceva",
			salt: "SOMErondomSTUFF",
			created: "2018-12-31T23:51:55Z"
		},
		"k2": {
			aes_key: "CEVAjmo91Jaj12kjbkPE89Rj89fceva1",
			salt: "SOMErondomSTUFF",
			created: "2018-12-31T23:51:55Z",
			compromised: true
		},
		//The "default" key is used for decryption when there is no prefix set on the encrypted data with the key index name
		"default": {
			aes_key: "CEVAjmo91Jaj12kjbkPE89Rj89fceva1",
			salt: "SOMErondomSTUFF",
			created: "2018-12-31T23:51:55Z"
		}
	}
};

let encryptionInstance = new Encryption(
	objEncryptionKeysConfig,
	Encryption.CIPHER_ALGORITHM_AES_CBC //Optional - Default: CIPHER_ALGORITHM_AES_CBC
);

let strEncryptedData = await encryptionInstance.encrypt(strTestFeedData/*, Encryption.CIPHER_ALGORITHM_AES_CBC, false*/);
let bufferEncryptedDataAES = await encryptionInstance.encryptAES(strTestFeedData/*, false*/);

let {keyIndex, cipherAlgorithm, encryptedDataWithoutMetadata} = encryptionInstance.getMetadataFromEncryptedData(strEncryptedData);

let objDecryptOptions = {
	skipHashValidation: false,
	allowCompromisedKeys: false, //Default: Decryption throws and error when using compromised keys.
	payloadBase64Encoded: true, //Default true.
	hasPayloadHash: true, //Default: hash is in the payload. If set to false, then skipHashValidation is set to true and hash is ignored.
	hasPayloadIV: true, //Default: IV is in the payload. If set to false, then customIV must be set as a non-empty string or Buffer and used for decryption.
	customIV: null, //It is used only when hasPayloadIV is set to false
	offsetBytesBeforePayloadParse: 0, //If set to a positive integer, then that number of bytes will be skipped from payload before parsing it (right AFTER base64 decode)
	parseMetadata: true //Default: true - specifies if decrypt should try to parse encrypted data metadata
};

let bufferDecryptedData = await encryptionInstance.decrypt(strEncryptedData, objDecryptOptions);

let strDecryptedData = bufferDecryptedData.toString();

assert(strTestFeedData === strDecryptedData, `Test failed. Expected ${JSON.stringify(strTestFeedData)} and got ${JSON.stringify(strDecryptedData)}`);
```

# TODO

* EncryptionException class