"use strict";

const forge = require("node-forge");
const assert = require("assert");

module.exports =
class Encryption
{
	constructor(objConfig, strCipherAlgorithm = Encryption.CIPHER_ALGORITHM_AES_CBC)
	{
		assert(typeof objConfig === "object", "Required config needs to be of type Object in Encryption instance constructor.");
		assert(typeof objConfig.keys === "object", "Required property config.keys needs to be of type Object in Encryption instance.");
		assert(typeof objConfig.active_index === "string", "Required property config.active_index needs to be of type string in Encryption instance.");

		this.cipherAlgorithm = strCipherAlgorithm;
		this.keyMap = objConfig.keys;
		this.activeKeyIndex = objConfig.active_index;
		this.partition = objConfig.keys_partition;
	}

	get partition() { return this._strPartition || null; }
	set partition(value)
	{
		if(
			!(
				typeof value === "undefined" 
				|| value === null
				|| (
					typeof value === "string" && value.trim().length > 0
				)
			)
		)
		{
			throw new TypeError(`Invalid partition value set. Expected "undefined", "null" or non-empty "string", but got value ${JSON.stringify(value)}.`);
		}
		
		this._strPartition = value;
	}

	/**
	 * @returns {boolean}
	 */
	hasPartition()
	{
		return typeof this.partition === "string" && this.partition.trim().length > 0;
	}


	/**
	 * @returns {string}
	 */
	get cipherAlgorithm() { return this._strCipherAlgorithm; }

	/**
	 * @param {string} value
	 */
	set cipherAlgorithm(value)
	{
		this.constructor.validateCipherAlgorithm(value);

		this._strCipherAlgorithm = value;
	}

	/**
	 * Returns algorithm simple name, without it's mode in lower case. Example: "AES-CBC" -> "aes"
	 *
	 * @returns {string}
	 */
	get cipherAlgorithmWithoutMode()
	{
		assert(typeof this.cipherAlgorithm === "string", "Invalid cipher algorithm");
		assert(this.cipherAlgorithm.includes("-"), "Invalid cipher algorithm. No mode set on it's name.");

		let strCipherAlgorithm = this.cipherAlgorithm.split("-")[0];
		assert(typeof strCipherAlgorithm === "string" && strCipherAlgorithm.trim().length > 0, "Invalid cipher algorithm without mode.");

		return strCipherAlgorithm.toLowerCase();
	}

	/**
	 * Returns the property name for the key value of the specific cipher algorithm used by this instance.
	 *
	 * @returns {string}
	 */
	get cipherAlgorithmKeyPropertyName()
	{
		return `${this.cipherAlgorithmWithoutMode}_key`;
	}

	/**
	 * Returns the property name for the key salt of the specific cipher algorithm used by this instance.
	 *
	 * @returns {string}
	 */
	get saltKeyPropertyName()
	{
		return "salt";
	}

	/**
	 * @returns {Object}
	 */
	get keyMap() { return this._objKeyMap; }

	/**
	 * @param {string} objKeys
	 *
	 * @throws {Error|TypeError}
	 */
	set keyMap(objKeys)
	{
		if(typeof objKeys !== "object" || objKeys === null)
		{
			throw new TypeError(`Invalid key map type in ${this.constructor.name}. Expected non-null "object", but got "${typeof objKeys}"${objKeys === null ? " null" : ""}`);
		}

		for(let strIndex in objKeys)
		{
			assert(typeof objKeys[strIndex] === "object", "keys[strIndex] needs to be of type Object.");
			assert(typeof objKeys[strIndex][this.cipherAlgorithmKeyPropertyName] === "string", `keys[strIndex].${this.cipherAlgorithmKeyPropertyName} needs to be of type string.`);
			assert(typeof objKeys[strIndex].salt === "string", "keys[strIndex].salt needs to be of type string.");
			if(typeof objKeys[strIndex].created !== "undefined")
			{
				assert(typeof objKeys[strIndex].created === "string", "keys[strIndex].created needs to be of type string if set.");
			}

			// Validate key length
			// https://www.npmjs.com/package/node-forge#ciphers-1
			// Note: a key size of 16 bytes will use AES-128, 24 => AES-192, 32 => AES-256
			// Otherwise a not so explicit error is thrown in the crypto library
			if(![16, 24, 32].includes(objKeys[strIndex][this.cipherAlgorithmKeyPropertyName].length))
			{
				throw new Error(`Invalid ${this.cipherAlgorithmKeyPropertyName} key with index ${JSON.stringify(strIndex)}. The ${this.cipherAlgorithmKeyPropertyName} key must have 16, 24 or 32 hex characters, but found ${objKeys[strIndex][this.cipherAlgorithmKeyPropertyName].length} characters.`);
			}
		}

		this._objKeyMap = objKeys;
	}

	/**
	 * @returns {string}
	 */
	get activeKeyIndex() { return this._strActiveKeyIndex; }

	/**
	 * @param {string} value
	 */
	set activeKeyIndex(value)
	{
		if(typeof value !== "string" || value.trim().length === 0)
		{
			throw new TypeError(`Invalid key index type in ${this.constructor.name}. Expected non-empty "string", but got "${JSON.stringify(value)}"`);
		}

		if(value.includes(this.constructor.ENCRYPTION_METADATA_DELIMITER))
		{
			throw new Error(`Invalid key index value. It cannot contain encryption metadata delimiter character '${this.constructor.ENCRYPTION_METADATA_DELIMITER}'.`);
		}

		if(typeof this.keyMap[value] !== "object")
		{
			throw new TypeError(`There is no key object data set for active key index ${JSON.stringify(value)}.`);
		}
		let objActiveKey = this.keyMap[value];

		assert(
			objActiveKey.compromised !== true,
			`Key with index name "${value}" cannot be set active key because it is "compromised".`
		);

		this._strActiveKeyIndex = value;
	}

	/**
	 * @returns {Object}
	 */
	get activeKey()
	{
		if(typeof this.keyMap[this.activeKeyIndex] !== "object")
		{
			throw new TypeError(`There is no key object data set for active key index ${JSON.stringify(this.activeKeyIndex)}.`);
		}

		assert(
			this.keyMap[this.activeKeyIndex].compromised !== true,
			`Active key with index name "${this.activeKeyIndex}" cannot be "compromised"`
		);

		return this.keyMap[this.activeKeyIndex];
	}

	/**
	 * Returns active key value.
	 *
	 * @returns {string}
	 */
	get activeKeyValue()
	{
		return this.activeKey[this.cipherAlgorithmKeyPropertyName];
	}

	/**
	 * Returns active key salt.
	 *
	 * @returns {string}
	 */
	get activeKeySalt()
	{
		if(typeof this.activeKey[this.saltKeyPropertyName] !== "string")
		{
			throw new Error(`Invalid key with index "${this.activeKeyIndex}". Salt property "${this.saltKeyPropertyName}" for key must be a "string", but "${typeof this.activeKey[this.saltKeyPropertyName]}" found.`);
		}

		return this.activeKey[this.saltKeyPropertyName];
	}

	/**
	 *
	 * @param {string} strKeyIndex
	 * @param {string} [bAllowCompromisedKey=false]
	 *
	 * @returns {string}
	 */
	getKeyValue(strKeyIndex, bAllowCompromisedKey = false)
	{
		assert(typeof strKeyIndex === "string", `Invalid key index in getKeyValue function. Expected "string", but got ${JSON.stringify(strKeyIndex)}.`);
		assert(this.keyMap.hasOwnProperty(strKeyIndex), `Keys object does not contain key index ${JSON.stringify(strKeyIndex)}.`);

		if(this.keyMap[strKeyIndex].compromised === true && bAllowCompromisedKey !== true)
		{
			throw new Error(`Cannot use compromised key with index name "${strKeyIndex}". For using compromised keys, "bAllowCompromisedKey" flag can be set to "true". But use it with care!`);
		}

		return this.keyMap[strKeyIndex][this.cipherAlgorithmKeyPropertyName];
	}

	/**
	 *
	 * @param {string} strKeyIndex
	 *
	 * @returns {string}
	 */
	getKeySalt(strKeyIndex)
	{
		assert(typeof strKeyIndex === "string", `Invalid key index in getKeyValue function. Expected "string", but got ${JSON.stringify(strKeyIndex)}.`);
		assert(this.keyMap.hasOwnProperty(strKeyIndex), `Keys object does not contain key index ${JSON.stringify(strKeyIndex)}.`);

		if(typeof this.keyMap[strKeyIndex][this.saltKeyPropertyName] !== "string")
		{
			throw new Error(`Invalid key with index "${strKeyIndex}". Salt property "${this.saltKeyPropertyName}" for key must be a "string", but "${typeof this.keyMap[strKeyIndex][this.saltKeyPropertyName]}" found.`);
		}

		return this.keyMap[strKeyIndex][this.saltKeyPropertyName].toString();
	}

	/**
	 * @returns {string}
	 */
	getDefaultKeyValue()
	{
		assert(
			this.keyMap.hasOwnProperty(this.constructor.ENCRYPTION_KEY_INDEX_DEFAULT),
			`Keys object does not contain default key index "${this.constructor.ENCRYPTION_KEY_INDEX_DEFAULT}". It is needed for decrypting data without key index prefix included.`
		);

		assert(
			this.keyMap[this.constructor.ENCRYPTION_KEY_INDEX_DEFAULT].compromised !== true,
			`Default key with index name "${this.constructor.ENCRYPTION_KEY_INDEX_DEFAULT}" cannot be set as "compromised"`
		);

		return this.keyMap[this.constructor.ENCRYPTION_KEY_INDEX_DEFAULT][this.cipherAlgorithmKeyPropertyName];
	}

	/**
	 * Gets a hash of data as IV.
	 *
	 * @param {string|Buffer} strDataToBeEncrypted
	 * @param {string} strKeyIndex
	 *
	 * @returns {Buffer}
	*/
	generateHexHashForData(strDataToBeEncrypted, strKeyIndex)
	{
		return (this.generateBufferHashForData(strDataToBeEncrypted, strKeyIndex)).toString("hex");
	}

	/**
	 * Gets a hash of data as IV. Adds the salt of the key index before hashing.
	 *
	 * @param {string|Buffer} mxDataToBeEncrypted
	 * @param {string} strKeyIndex
	 *
	 * @returns {Buffer}
	*/
	generateBufferHashForData(mxDataToBeEncrypted, strKeyIndex)
	{
		if(typeof mxDataToBeEncrypted === "undefined")
		{
			throw new Error("Undefined data for generating hash buffer for data.");
		}

		if(mxDataToBeEncrypted === null)
		{
			throw new Error("Null data for generating hash buffer for data.");
		}

		let strDataToBeEncryptedWithSalt = mxDataToBeEncrypted;

		if(Buffer.isBuffer(mxDataToBeEncrypted))
		{
			strDataToBeEncryptedWithSalt = mxDataToBeEncrypted.toString();
		}
		else if(typeof mxDataToBeEncrypted !== "string")
		{
			throw new Error("strDataToBeEncrypted must be string or Buffer.");
		}

		strDataToBeEncryptedWithSalt = mxDataToBeEncrypted + this.getKeySalt(strKeyIndex);

		let hmac = forge.hmac.create();
		hmac.start("md5", this.getKeyValue(strKeyIndex), true);
		hmac.update(strDataToBeEncryptedWithSalt);

		const bufferHash = Buffer.from(hmac.digest().getBytes(), "binary");

		assert(
			bufferHash.byteLength === this.constructor.ENCRYPTION_HASH_LENGTH_BYTES,
			`Invalid hash generated. Expected a Buffer with ${this.constructor.ENCRYPTION_HASH_LENGTH_BYTES} bytes, but got ${bufferHash.byteLength} bytes.`
		);

		return bufferHash;
	}

	/**
	 * @private
	 * @param {string} strKeyIndex 
	 * 
	 * @returns {string}
	 */
	_getKeyNameWithPartition(strKeyIndex)
	{
		if(this.hasPartition())
		{
			return this.partition + this.constructor.ENCRYPTION_PARTITION_DELIMITER + strKeyIndex;
		}

		return strKeyIndex;
	}


	/**
	 * Returns a lower case hex string, prefixed with information.
	 * 
	 * @param {string|Buffer} strDataToBeEncrypted
	 * @param {string|null} [strAlgorithm="sha256"]
	 * @param {string|null} [strKeyIndex=null]
	 *
	 * @returns {string}
	 */
	async hash(strDataToBeHashed, strAlgorithm = "hmac-sha256", strKeyIndex = null)
	{
		if(!this.constructor.allowedHashAlgorithms.includes(strAlgorithm))
		{
			throw new Error(`Unknown hash algorithm ${strAlgorithm}. Maybe you mean one of these: ${JSON.stringify(this.constructor.allowedHashAlgorithms)}`);
		}

		if(strKeyIndex === null)
		{
			strKeyIndex = this.activeKeyIndex;
		}

		if(strAlgorithm.startsWith("hmac-"))
		{
			const hmac = forge.hmac.create();
			hmac.start(strAlgorithm.substr(5), this.getKeySalt(strKeyIndex));
			hmac.update(strDataToBeHashed);

			return `${strKeyIndex}|${strAlgorithm}|${hmac.digest().toHex().toLowerCase()}`;
		}
		else
		{
			let md = forge.md[strAlgorithm].create();
			md.update(strDataToBeHashed, "utf8");

			return `${strKeyIndex}|${strAlgorithm}|${md.digest().toHex().toLowerCase()}`;
		}
	}


	/**
	 * @param {string} strHashData 
	 * 
	 * @returns {boolean}
	 */
	isHash(strHashData)
	{
		return this.constructor.allowedHashAlgorithms.includes(strHashData.split("|")[1]);
	}


	/**
	 * It uses only the active key for encryption.
	 * Format: "KEY_INDEX_NAME|ALGORITHM_NAME_LOWER_CASE|" + base64(iv + data_hash + encrypted_data)
	 * 
	 * iv - 16 bytes
	 * hash - 16 bytes
	 *
	 * @param {string|Buffer} strDataToBeEncrypted
	 * @param {string|null} [strCipherAlgorithm=null] if null, it uses this instance encryptionAlgorithm
	 * @param {boolean} [bDisableDecryptValidation=false]
	 * @param {Buffer|null} [bufferIV=null]
	 *
	 * @returns {string}
	 */
	async encrypt(strDataToBeEncrypted, strCipherAlgorithm = null, bDisableDecryptValidation = false, bufferIV = null)
	{
		if(typeof strDataToBeEncrypted === "undefined")
		{
			throw new Error("Undefined data to be encrypted.");
		}

		if(strDataToBeEncrypted === null)
		{
			throw new Error("Null data to be encrypted.");
		}

		let strEncryptionCipherAlgorithm = this.cipherAlgorithm;
		if(typeof strCipherAlgorithm === "string")
		{
			strEncryptionCipherAlgorithm = strCipherAlgorithm;
		}

		bufferIV = bufferIV ? bufferIV : await this.constructor.generateRandomBytes(this.constructor.ENCRYPTION_IV_LENGTH_BYTES);
		const bufferHash = this.generateBufferHashForData(strDataToBeEncrypted, this.activeKeyIndex);

		let cipher = forge.cipher.createCipher(strEncryptionCipherAlgorithm, this.activeKeyValue);
		cipher.start({iv: forge.util.createBuffer(bufferIV)});
		cipher.update(forge.util.createBuffer(strDataToBeEncrypted));
		cipher.finish();

		let strMetadata = [
			this._getKeyNameWithPartition(this.activeKeyIndex),
			this.cipherAlgorithm.toLowerCase()
		].join(this.constructor.ENCRYPTION_METADATA_DELIMITER);

		let strEncryptedData = strMetadata +
			this.constructor.ENCRYPTION_METADATA_DELIMITER +
			Buffer.concat(
				[
					bufferIV,
					bufferHash,
					Buffer.from(cipher.output.getBytes(), "binary")
				]
			).toString("base64");

		if(!bDisableDecryptValidation)
		{
			let bufferDecryptedData = await this.decrypt(strEncryptedData);
			let mdDecryptedData = forge.md.md5.create();
			mdDecryptedData.update(bufferDecryptedData);

			let mdOriginalData = forge.md.md5.create();
			mdOriginalData.update(strDataToBeEncrypted);

			if(mdDecryptedData.digest().toHex() !== mdOriginalData.digest().toHex())
			{
				throw new Error("Encrypted data failed decryption test.");
			}
		}

		return strEncryptedData;
	}


	/**
	 * Format: "KEY_INDEX_NAME|ALGORITHM_NAME_LOWER_CASE|" + base64(iv + data_hash + encrypted_data)
	 * 
	 * iv - 16 bytes
	 * hash - 16 bytes
	 *
	 * @param {string | Buffer} encryptedData
	 * @param {{skipHashValidation: boolean = false, allowCompromisedKeys: boolean = false, payloadBase64Encoded: boolean = true, hasPayloadHash: boolean = true, hasPayloadIV: boolean = true, customIV: string | Buffer = null}} [objDecryptOptions={}] Options
	 *
	 * @returns {Buffer}
	 */
	async decrypt(
		encryptedData,
		{
			skipHashValidation: bSkipHashSignatureValidation = false,
			allowCompromisedKeys: bAllowCompromisedKeys = false,
			payloadBase64Encoded: bIsPayloadBase64Encoded = true,
			hasPayloadHash: bHasPayloadHash = true,
			hasPayloadIV: bHasPayloadIV = true,
			customIV: mxCustomIV = null,
			offsetBytesBeforePayloadParse: nOffsetBytesBeforePayloadParse = 0,
			parseMetadata: bParseMetadata = true
		} = {}
	)
	{
		//Validation
		if(
			typeof encryptedData === "undefined"
			|| encryptedData === null
		)
		{
			throw new TypeError(`Invalid encrypted data for decrypt: ${JSON.stringify(encryptedData)}`);
		}

		assert(typeof bSkipHashSignatureValidation === "boolean", `Invalid option skipHashValidation. Expected type "boolean", but got value: ${JSON.stringify(bSkipHashSignatureValidation)}.`);
		assert(typeof bAllowCompromisedKeys === "boolean", `Invalid option allowCompromisedKeys. Expected type "boolean", but got value: ${JSON.stringify(bAllowCompromisedKeys)}.`);
		assert(typeof bHasPayloadHash === "boolean", `Invalid option hasPayloadHash. Expected type "boolean", but got value: ${JSON.stringify(bHasPayloadHash)}.`);
		assert(typeof bHasPayloadIV === "boolean", `Invalid option hasPayloadIV. Expected type "boolean", but got value: ${JSON.stringify(bHasPayloadIV)}.`);
		assert(typeof bIsPayloadBase64Encoded === "boolean", `Invalid option payloadBase64Encoded. Expected type "boolean", but got value: ${JSON.stringify(bIsPayloadBase64Encoded)}.`);
		assert(typeof bParseMetadata === "boolean", `Invalid option parseMetadata. Expected type "boolean", but got value: ${JSON.stringify(bParseMetadata)}.`);
		assert(Number.isInteger(nOffsetBytesBeforePayloadParse), `Invalid option offsetBytesBeforePayloadParse. Expected type "Integer", but got value: ${JSON.stringify(nOffsetBytesBeforePayloadParse)}.`);
		
		if(!bHasPayloadHash)
		{
			bSkipHashSignatureValidation = true;
		}
		if(!bHasPayloadIV)
		{
			if(typeof mxCustomIV === "string" && mxCustomIV.trim().length > 0)
			{
				mxCustomIV = Buffer.from(mxCustomIV);
			}
			else if(!Buffer.isBuffer(mxCustomIV))
			{
				throw new TypeError(`Invalid custom IV. If hasPayloadIV option is set to false, a custom IV is expected to be given as a non-empty "string" or a "Buffer". Custom IV value got: ${JSON.stringify(mxCustomIV)}`);
			}
		}

		//Metadata
		let {keyIndex: strKeyIndex, cipherAlgorithm: strCipherAlgorithm, encryptedDataWithoutMetadata: bufferEncryptedDataWithoutMetadata} = (
			bParseMetadata === false 
				? {keyIndex: null, cipherAlgorithm: null, encryptedDataWithoutMetadata: encryptedData}
				: this.getMetadataFromEncryptedData(encryptedData)
		);

		let strDecryptionKeyValue;
		if(typeof strKeyIndex === "undefined" || strKeyIndex === null)
		{
			strDecryptionKeyValue = this.getDefaultKeyValue();
			strKeyIndex = this.constructor.ENCRYPTION_KEY_INDEX_DEFAULT;
		}
		else
		{
			strDecryptionKeyValue = this.getKeyValue(strKeyIndex, bAllowCompromisedKeys);
		}

		if(typeof strCipherAlgorithm === "undefined" || strCipherAlgorithm === null)
		{
			strCipherAlgorithm = this.cipherAlgorithm;
		}

		assert(typeof strDecryptionKeyValue === "string", `There is no key value set for decrypt for key index ${strKeyIndex} -> ${strDecryptionKeyValue}.`);
		assert(typeof strCipherAlgorithm === "string", "There is no cipher algorithm name set for decryption.");

		//Payload
		//convert from base64 to binary
		let bufferFullEncryptedData = bIsPayloadBase64Encoded ? Buffer.from(bufferEncryptedDataWithoutMetadata.toString(), "base64") : bufferEncryptedDataWithoutMetadata;

		if(nOffsetBytesBeforePayloadParse > 0)
		{
			bufferFullEncryptedData = bufferFullEncryptedData.slice(nOffsetBytesBeforePayloadParse);
		}

		let nIVLength = bHasPayloadIV ? this.constructor.ENCRYPTION_IV_LENGTH_BYTES : 0;
		let nHashLength = bHasPayloadHash ? this.constructor.ENCRYPTION_HASH_LENGTH_BYTES : 0;

		let {iv: bufferIV, hash: bufferHash, encryptedData: bufferCipherData} = this.parsePayload(bufferFullEncryptedData, nIVLength, nHashLength);

		assert(!bHasPayloadIV || (bHasPayloadIV && bufferIV), "Invalid IV in encrypted data after payload parse.");
		assert(!bHasPayloadHash || (bHasPayloadHash && bufferHash), "Invalid Hash in encrypted data after payload parse.");

		if(!bHasPayloadIV)
		{
			bufferIV = mxCustomIV;
		}

		//Decrypt
		let decipher = forge.cipher.createDecipher(strCipherAlgorithm, strDecryptionKeyValue);

		decipher.start({iv: forge.util.createBuffer(bufferIV)});
		decipher.update(forge.util.createBuffer(bufferCipherData));
		decipher.finish();

		let bufferDecryptedData = Buffer.from(decipher.output.getBytes(), "binary");

		//Validation
		if(!bSkipHashSignatureValidation)
		{
			const bufferHashForDecryptedData = this.generateBufferHashForData(bufferDecryptedData, strKeyIndex);
			if(!bufferHashForDecryptedData.equals(bufferHash))
			{
				throw new Error(`Invalid encrypted data. Failed hash signature validation. Payload hash (hex): ${bufferHash.toString("hex")} != Decrypted data hash (hex): ${bufferHashForDecryptedData.toString("hex")}`);
			}
		}

		return bufferDecryptedData;
	}

	/**
	 * If in the payload is no IV or no Hash, set nIVLengthBytes and nHashLengthBytes, respectively, to 0. 
	 * If set to 0, the corresponding response property will be null.
	 * 
	 * @param {string | Buffer} mxEncryptedDataPayload
	 * @param {Integer} [nIVLengthBytes=Encryption.ENCRYPTION_IV_LENGTH_BYTES_DEFAULT]
	 * @param {Integer} [nHashLengthBytes=Encryption.ENCRYPTION_HASH_LENGTH_BYTES]
	 *
	 * @returns {{iv: Buffer, hash: Buffer, encryptedData: Buffer}}
	 */
	parsePayload(mxEncryptedDataPayload, nIVLengthBytes = Encryption.ENCRYPTION_IV_LENGTH_BYTES, nHashLengthBytes = Encryption.ENCRYPTION_HASH_LENGTH_BYTES)
	{
		if(
			typeof mxEncryptedDataPayload === "undefined"
			|| mxEncryptedDataPayload === null
		)
		{
			throw new TypeError(`Invalid encrypted data payload for parsing: ${JSON.stringify(mxEncryptedDataPayload)}`);
		}

		if(nIVLengthBytes < 0)
		{
			throw new Error(`Invalid IV length. Expected positive integer, but got ${JSON.stringify(nIVLengthBytes)}`);
		}

		if(nHashLengthBytes < 0)
		{
			throw new Error(`Invalid Hash length. Expected positive integer, but got ${JSON.stringify(nIVLengthBytes)}`);
		}

		let bufferEncryptedDataPayload;

		//Convert to Buffer
		if(!(Buffer.isBuffer(mxEncryptedDataPayload)))
		{
			bufferEncryptedDataPayload = Buffer.from(mxEncryptedDataPayload, "binary");
		}
		else
		{
			bufferEncryptedDataPayload = mxEncryptedDataPayload;
		}

		if(nIVLengthBytes === 0 && nHashLengthBytes === 0)
		{
			return {
				iv: null,
				hash: null,
				encryptedData: bufferEncryptedDataPayload
			};
		}

		assert(
			bufferEncryptedDataPayload.byteLength > nIVLengthBytes + nHashLengthBytes,
			(
				`Invalid encrypted data. Encrypted data length must be greater than the sum of IV length ${nIVLengthBytes} and Hash length ${nHashLengthBytes}.` +
				`Found length of ${bufferEncryptedDataPayload.byteLength} bytes. It is possible that the encrypted data is not encrypted with IV and Hash included or another encryption algorithm was used or data was corrupted.`
			)
		);

		return {
			iv: nIVLengthBytes > 0 ? bufferEncryptedDataPayload.slice(0, nIVLengthBytes) : null,
			hash: nHashLengthBytes > 0 ? bufferEncryptedDataPayload.slice(nIVLengthBytes, nIVLengthBytes + nHashLengthBytes) : null,
			encryptedData: bufferEncryptedDataPayload.slice(nIVLengthBytes + nHashLengthBytes)
		};
	}


	/**
	 * @param {string} strHashData
	 *
	 * @returns {{keyIndex: string, hashAlgorithm: string, hash: string}}
	 */
	getMetadataFromHashData(strHashData)
	{
		assert(typeof strHashData === "string", "strHashData was suposed to be a string.");

		const arrParts = strHashData.split("|");

		return {
			keyIndex: arrParts[0],
			hashAlgorithm: arrParts[1],
			hash: arrParts[2]
		};
	}


	/**
	 *
	 * @param {string | Buffer} encryptedData
	 *
	 * @returns {{keyIndex: string|null, cipherAlgorithm: string|null, encryptedDataWithoutMetadata: Buffer}}
	 */
	getMetadataFromEncryptedData(encryptedData)
	{
		if(
			typeof encryptedData === "undefined"
			|| encryptedData === null
		)
		{
			throw new TypeError(`Invalid encrypted data for getting key index ${JSON.stringify(encryptedData)}`);
		}

		let bufferEncryptedData;

		if(!(Buffer.isBuffer(encryptedData)))
		{
			bufferEncryptedData = Buffer.from(encryptedData, "binary");
		}
		else
		{
			bufferEncryptedData = encryptedData;
		}

		//Key Index
		let nKeyIndexDelimiterIndex = bufferEncryptedData.indexOf(this.constructor.ENCRYPTION_METADATA_DELIMITER, 0);
		assert(nKeyIndexDelimiterIndex + 1 < bufferEncryptedData.length, `Invalid delimiter index ${nKeyIndexDelimiterIndex}. It is greater than encrypted data length of ${bufferEncryptedData.length}.`);

		//No delimiter
		if(nKeyIndexDelimiterIndex < 0)
		{
			// throw new Error(`Invalid encrypted data for getting key index. It does not contain key index delimiter '${this.constructor.ENCRYPTION_METADATA_DELIMITER}'.`);
			return {
				keyIndex: null,
				cipherAlgorithm: null,
				encryptedDataWithoutMetadata: bufferEncryptedData
			};
		}

		let strKeyIndex = bufferEncryptedData.toString("utf8", 0, nKeyIndexDelimiterIndex);
		if(this.hasPartition())
		{
			let strPrefix = this.partition + this.constructor.ENCRYPTION_PARTITION_DELIMITER;
			if(!strKeyIndex.startsWith(strPrefix))
			{
				throw new Error(`Key index name got from encrypted data does not start with partition prefix set in config object, ${JSON.stringify(this.partition)} and the partition delimiter '${this.constructor.ENCRYPTION_PARTITION_DELIMITER}'.`);
			}

			strKeyIndex = strKeyIndex.substr(strPrefix.length);
		}

		if(!this.keyMap.hasOwnProperty(strKeyIndex))
		{
			//key index missing, but it is a cipher algorithm as the first metadata
			if(this.constructor.allowedCipherAlgorithms.includes(strKeyIndex.toUpperCase()))
			{
				throw new Error(`Invalid metadata. Key index is missing from encrypted data or not set first, but cipher algorithm "${strKeyIndex}" was set as first metadata item.`);
			}
			else
			{
				throw new Error(`Key index got from encrypted data, "${strKeyIndex}", is not set in keys map.`);
			}
		}

		//Cipher Algorithm
		const bufferEncryptedDataWithoutKeyIndex = bufferEncryptedData.slice(nKeyIndexDelimiterIndex + 1);
		let nCipherAlgorithmDelimiterIndex = bufferEncryptedDataWithoutKeyIndex.indexOf(this.constructor.ENCRYPTION_METADATA_DELIMITER);
		assert(nCipherAlgorithmDelimiterIndex + 1 < bufferEncryptedDataWithoutKeyIndex.length, `Invalid delimiter index ${nCipherAlgorithmDelimiterIndex}. It is greater than encrypted data length of ${bufferEncryptedDataWithoutKeyIndex.length}.`);

		if(nCipherAlgorithmDelimiterIndex < 0)
		{
			// throw new Error(`Invalid encrypted data for getting cipher algorithm. It does not contain metadata delimiter '${this.constructor.ENCRYPTION_METADATA_DELIMITER}'.`);
			return {
				keyIndex: strKeyIndex,
				cipherAlgorithm: null,
				encryptedDataWithoutMetadata: bufferEncryptedDataWithoutKeyIndex
			};
		}

		const strCipherAlgorithm = bufferEncryptedDataWithoutKeyIndex.toString("utf8", 0, nCipherAlgorithmDelimiterIndex).toUpperCase();
		this.constructor.validateCipherAlgorithm(strCipherAlgorithm)
		
		return {
			keyIndex: strKeyIndex,
			cipherAlgorithm: strCipherAlgorithm,
			encryptedDataWithoutMetadata: bufferEncryptedDataWithoutKeyIndex.slice(nCipherAlgorithmDelimiterIndex + 1)
		};
	}

	/**
	 *
	 * @param {string | Buffer} encryptedData
	 * @param {Object} objDecryptOptions
	 * @param {string} strEncryptionAlgorithm
	 *
	 * @returns {Buffer}
	 */
	async rotateToActiveKey(
		encryptedData,
		objDecryptOptions = undefined,
		strEncryptionAlgorithm = Encryption.CIPHER_ALGORITHM_AES_CBC
	)
	{
		return this.rotateKey(this.activeKeyIndex, this.activeKey, encryptedData, strEncryptionAlgorithm);
	}

	/**
	 *
	 * @param {string} strNewKeyIndex
	 * @param {Object} objNewKeyData
	 * @param {string | Buffer} encryptedData
	 * @param {Object} objDecryptOptions
	 * @param {string} [strEncryptionAlgorithm=Encryption.CIPHER_ALGORITHM_AES_CBC]
	 *
	 * @returns {Buffer}
	 */
	async rotateKey(
		strNewKeyIndex,
		objNewKeyData,
		encryptedData,
		objDecryptOptions = undefined,
		strEncryptionAlgorithm = Encryption.CIPHER_ALGORITHM_AES_CBC
	)
	{
		let encryptionInstance = new Encryption(
			{
				keys: Object.assign(
					this.keyMap,
					{
						[strNewKeyIndex]: objNewKeyData
					}
				),
				active_index: strNewKeyIndex
			},
			strEncryptionAlgorithm
		);
		if(
			typeof encryptedData === "undefined"
			|| encryptedData === null
		)
		{
			throw new TypeError(`Invalid encrypted data for rotate key ${JSON.stringify(encryptedData)}`);
		}

		let decryptedData = await encryptionInstance.decrypt(encryptedData, objDecryptOptions);

		let newEncryptedData = await encryptionInstance.encrypt(decryptedData, strEncryptionAlgorithm, /*bDisableDecryptValidation*/ false);

		assert(newEncryptedData.includes(strNewKeyIndex), "Rotate key result does not include the new key index.");

		return newEncryptedData;
	}

	/**
	 * 
	 * @param {string | Buffer} strDataToBeEncrypted 
	 * @param {string} strKeyValue 
	 * @param {Buffer} [bufferIV=null]
	 * @param {string} [strCipherAlgorithm=Encryption.CIPHER_ALGORITHM_AES_CBC]
	 * 
	 * @returns {{iv: Buffer, encryptedData: Buffer}}
	 */
	static async encryptRaw(strDataToBeEncrypted, strKeyValue, bufferIV = null, strCipherAlgorithm = Encryption.CIPHER_ALGORITHM_AES_CBC)
	{
		if(bufferIV === null)
		{
			bufferIV = await this.generateRandomBytes(this.ENCRYPTION_IV_LENGTH_BYTES);
		}

		let cipher = forge.cipher.createCipher(strCipherAlgorithm, strKeyValue);
		cipher.start({iv: forge.util.createBuffer(bufferIV)});
		cipher.update(forge.util.createBuffer(strDataToBeEncrypted));
		cipher.finish();

		return {
			iv: bufferIV,
			encryptedData: Buffer.from(cipher.output.getBytes(), "binary")
		};
	}

	static validateCipherAlgorithm(strCipherAlgorithm)
	{
		if(typeof strCipherAlgorithm !== "string" || strCipherAlgorithm.trim().length === 0)
		{
			throw new TypeError(`Invalid type for cipher algorithm. Expected non-empty "string", but got the value ${JSON.stringify(strCipherAlgorithm)}.`);
		}

		if(!this.allowedCipherAlgorithms.includes(strCipherAlgorithm))
		{
			throw new Error(`Invalid cipher algorithm ${JSON.stringify(strCipherAlgorithm)} in ${this.name}. Allowed ones are: ${this.allowedCipherAlgorithms}`);
		}
	}

	/**
	 * 
	 * @param {Integer} nLength 
	 * @param {string} [strEncoding="ascii"] 
	 * 
	 * @returns {string}
	 */
	static async generateRandomString(nLength, strEncoding = "ascii")
	{
		return (await this.generateRandomBytes(nLength)).toString(strEncoding);
	}

	/**
	 * @param {Integer} nLength
	 * 
	 * @returns {Buffer}
	*/
	static async generateRandomBytes(nLength)
	{
		let bufferIV = Buffer.from(forge.random.getBytesSync(nLength), "binary");
		assert(
			bufferIV.byteLength === nLength,
			`Wrong IV generated of bytes length ${bufferIV.byteLength}. Expected ${nLength}.`
		);

		return bufferIV;
	}

	/**
	 * @returns {Array}
	 */
	static get allowedCipherAlgorithms()
	{
		return [
			this.CIPHER_ALGORITHM_AES_CBC
		];
	}


	/**
	 * @returns {Array<string>}
	 */
	static get allowedHashAlgorithms()
	{
		return ["hmac-sha256", "hmac-sha1", "hmac-sha512", "hmac-sha384", "hmac-md5", "sha256", "sha1", "sha512", "sha384", "md5"];
	}


	static get CIPHER_ALGORITHM_AES_CBC() { return "AES-CBC"; }

	static get ENCRYPTION_METADATA_DELIMITER() { return "|"; }
	static get ENCRYPTION_PARTITION_DELIMITER() { return "_"; }
	static get ENCRYPTION_METADATA_DELIMITER_COUNT() { return 2; }

	static get ENCRYPTION_KEY_INDEX_DEFAULT() { return "default"; }

	static get ENCRYPTION_IV_LENGTH_BYTES() { return 16; }
	static get ENCRYPTION_HASH_LENGTH_BYTES() { return 16; }
};
