"use strict";

const assert = require("assert");

module.exports =
class BaseTest
{
	constructor(objContext = {})
	{
		this.context = objContext;
	}

	get context() { return this._objContext; }
	set context(value)
	{
		assert(typeof value === "object", `Invalid context object ${JSON.stringify(value)} for ${this.constructor.name}.`);
		this._objContext = value;
	}

	/**
	 * Used when modifying context object only in test function.
	 * 
	 * @returns {Object}
	 */
	getContextClone()
	{
		return JSON.parse(JSON.stringify(this.context));
	}

	//https://stackoverflow.com/questions/9781218/how-to-change-node-jss-console-font-color
	async runAllTests()
	{
		this.logInfo(`----- RUNNING ${this.constructor.name} -----`);
		try
		{
			const nStartTimeMs = (new Date()).getTime();
			await this.runTests();

			const nFinishTimeMs = (new Date()).getTime();
			this.logSuccess(`----- OK ${this.constructor.name} ${nFinishTimeMs - nStartTimeMs} ms -----`);
		}
		catch(error)
		{
			this.logError(`${this.constructor.name} failed with error:`, error);
			throw error;
		}
	}

	async runTests()
	{
		throw new Error(`Method "runTests" must be implemented in class ${this.constructor.name}.`)
	}

	static get CONTEXT_KEY() 
	{
		return this.name.toLowerCase().replace("test", "");
	}

	logError(message)
	{
		console.error.apply(this, ["\x1b[31m%s\x1b[0m", message]);
	}

	logSuccess(message)
	{
		console.log.apply(this, ["\x1b[32m%s\x1b[0m", message]);
	}

	logWarning(message)
	{
		console.warn.apply(this, ["\x1b[33m%s\x1b[0m", message]);
	}

	logInfo(message)
	{
		console.info.apply(this, ["\x1b[34m%s\x1b[0m", message]);
	}
};
