"use strict";

const BaseTest = require("./Base/BaseTest");
const assert = require("assert");
const Encryption = require("../index");

module.exports =
class EncryptionTest extends BaseTest
{
	async runTests()
	{
		await this.testSimpleEncryption();
		await this.testBufferEncryption();
		await this.testRotate();
		await this.testRotateToActiveKey();
		await this.testNoMetadata();
		await this.testNoKeyIndex();
		await this.testNoCipherAlgorithm();
		await this.testWrongKeyIndex();
		await this.testWrongCipherAlgorithmInMetadata();
		await this.testAlteredEncryptedData();
		await this.testCompromisedActiveKey();
		await this.testCompromisedDefaultKey();
		await this.testCompromisedKeyDecryption();
		await this.testDecryptBadOptions();
		await this.testEncryptRawAndCustomDecrypt();
		await this.testDecryptOffsetBytes();
		await this.testKeysPartition();
		await this.testMultipleRandomEncryption();
		await this.testHMAC();
		await this.testHash();
	}

	
	async testHMAC()
	{
		const objEncryptionKeysConfig = this.context;
		const encryptionInstance = new Encryption.Encryption(objEncryptionKeysConfig);
		const strMetaDataAndHash = await encryptionInstance.hash("abcdefg");

		assert(/^[a-z]+[0-9]+\|hmac-sha256\|[0-9a-f]{64}$/.test(strMetaDataAndHash), strMetaDataAndHash);

		assert(encryptionInstance.isHash(strMetaDataAndHash));

		const objMetadata = encryptionInstance.getMetadataFromHashData(strMetaDataAndHash);
		assert([objMetadata.keyIndex, objMetadata.hashAlgorithm, objMetadata.hash].join("|") === strMetaDataAndHash, "Failed hash metadata test.");

		this.logSuccess("[OK] testHMAC");
	}


	async testHash()
	{
		const objEncryptionKeysConfig = this.context;
		const encryptionInstance = new Encryption.Encryption(objEncryptionKeysConfig);
		const strMetaDataAndHash = await encryptionInstance.hash("abcdefg", "md5");

		assert(encryptionInstance.isHash(strMetaDataAndHash));

		assert(/^[a-z]+[0-9]+\|md5\|[0-9a-f]{32}$/.test(strMetaDataAndHash), strMetaDataAndHash);

		const objMetadata = encryptionInstance.getMetadataFromHashData(strMetaDataAndHash);
		assert([objMetadata.keyIndex, objMetadata.hashAlgorithm, objMetadata.hash].join("|") === strMetaDataAndHash, "Failed hash metadata test.");

		this.logSuccess("[OK] testHash");
	}


	async testSimpleEncryption()
	{
		console.log("Running testSimpleEncryption...");

		let {encryptionInstance, strEncryptedData} = await this.generateEncryptedDataDefault();

		let bufferDecryptedData = await encryptionInstance.decrypt(strEncryptedData);

		let strDecryptedData = bufferDecryptedData.toString();

		assert(this.constructor.FEED_DATA_DEFAULT === strDecryptedData, `Test failed: testSimpleEncryption. Expected ${JSON.stringify(this.constructor.FEED_DATA_DEFAULT)} and got ${JSON.stringify(strDecryptedData)}`);

		this.logSuccess("[OK] testSimpleEncryption");
	}

	async testBufferEncryption()
	{
		console.log("Running testBufferEncryption...");

		const bufferTestFeedData = await Encryption.Encryption.generateRandomBytes(100);

		let objEncryptionKeysConfig = this.context;

		let encryptionInstance = new Encryption.Encryption(objEncryptionKeysConfig);

		let strEncryptedData = await encryptionInstance.encrypt(bufferTestFeedData);
		this.validateEncryptedDataFormat(strEncryptedData);

		let bufferDecryptedData = await encryptionInstance.decrypt(strEncryptedData);

		assert(bufferDecryptedData.equals(bufferTestFeedData), "Test failed: testBufferEncryption. Decrypted Buffer is not equal to test Buffer.");

		this.logSuccess("[OK] testBufferEncryption");
	}

	async testRotate()
	{
		console.log("Running testRotate...");

		const strTestFeedData = this.constructor.FEED_DATA_DEFAULT;
		const strNewKeyIndex = "k12345";
		const strNewKeyValue = await Encryption.Encryption.generateRandomString(32);
		const salt = await Encryption.Encryption.generateRandomString(32);

		let objEncryptionKeysConfig = this.getContextClone();
		objEncryptionKeysConfig.keys[strNewKeyIndex] = {
			aes_key: strNewKeyValue,
			salt: salt,
			created: "2018-12-31T23:51:55Z"
		};

		let {encryptionInstance, strEncryptedData} = await this.generateEncryptedDataDefault(objEncryptionKeysConfig);

		let newBufferEncryptedData = await encryptionInstance.rotateKey(
			strNewKeyIndex,
			objEncryptionKeysConfig.keys[strNewKeyIndex],
			strEncryptedData,
			Encryption.Encryption.CIPHER_ALGORITHM_AES_CBC,
		);

		let bufferDecryptedData = await encryptionInstance.decrypt(newBufferEncryptedData);

		let strDecryptedData = bufferDecryptedData.toString("utf8");

		assert(strTestFeedData === strDecryptedData, `Test failed: testRotate. Expected ${JSON.stringify(strTestFeedData)} and got ${JSON.stringify(strDecryptedData)}`);

		this.logSuccess("[OK] testRotate");
	}

	async testRotateToActiveKey()
	{
		console.log("Running testRotateToActiveKey...");

		const strTestFeedData = this.constructor.FEED_DATA_DEFAULT;
		const strNewKeyIndex = "k12345";
		const strNewKeyValue = await Encryption.Encryption.generateRandomString(32);
		const salt = await Encryption.Encryption.generateRandomString(32);

		let objEncryptionKeysConfig = this.context;

		let encryptionInstanceForEncrypt = new Encryption.Encryption(
			{
				keys: Object.assign(
					objEncryptionKeysConfig.keys,
					{
						[strNewKeyIndex]: {
							aes_key: strNewKeyValue,
							salt: salt,
							created: "2018-12-31T23:51:55Z"
						}
					}
				),
				active_index: strNewKeyIndex
			}
		);

		let strEncryptedData = await encryptionInstanceForEncrypt.encrypt(strTestFeedData);
		this.validateEncryptedDataFormat(strEncryptedData);

		let encryptionInstance = new Encryption.Encryption(
			objEncryptionKeysConfig,
			Encryption.Encryption.CIPHER_ALGORITHM_AES_CBC
		);

		let newBufferEncryptedData = await encryptionInstance.rotateToActiveKey(strEncryptedData);

		let bufferDecryptedData = await encryptionInstance.decrypt(newBufferEncryptedData);

		let strDecryptedData = bufferDecryptedData.toString("utf8");

		assert(strTestFeedData === strDecryptedData, `Test failed: testRotateToActiveKey. Expected ${JSON.stringify(strTestFeedData)} and got ${JSON.stringify(strDecryptedData)}`);

		this.logSuccess("[OK] testRotateToActiveKey");
	}

	/**
	 * Removes all metadata from encrypted data.
	 * Expects for decryption to work with default key and default algorithm.
	 */
	async testNoMetadata()
	{
		console.log("Running testNoMetadata...");

		const strTestFeedData = this.constructor.FEED_DATA_DEFAULT;

		let objEncryptionKeysConfig = this.getContextClone();
		objEncryptionKeysConfig.keys[Encryption.Encryption.ENCRYPTION_KEY_INDEX_DEFAULT] = objEncryptionKeysConfig.keys[objEncryptionKeysConfig.active_index];

		let {encryptionInstance, strEncryptedData} = await this.generateEncryptedDataDefault(objEncryptionKeysConfig);

		let strEncryptedDataWithoutMetadata = this.removeAllMetadata(strEncryptedData);

		let bufferDecryptedData = await encryptionInstance.decrypt(strEncryptedDataWithoutMetadata);
		let strDecryptedData = bufferDecryptedData.toString();

		assert(strTestFeedData === strDecryptedData, `Test failed: testNoMetadata. Expected ${JSON.stringify(strTestFeedData)} and got ${JSON.stringify(strDecryptedData)}`);

		this.logSuccess("[OK] testNoMetadata");
	}

	/**
	 * Removes only the first item from encrypted data metadata.
	 * Expects an error to be thrown because the first item, instead of a key index, it will be the cipher algorithm.
	 */
	async testNoKeyIndex()
	{
		console.log("Running testNoKeyIndex...");

		let objEncryptionKeysConfig = this.getContextClone();
		objEncryptionKeysConfig.keys[Encryption.Encryption.ENCRYPTION_KEY_INDEX_DEFAULT] = objEncryptionKeysConfig.keys[objEncryptionKeysConfig.active_index];

		let {encryptionInstance, strEncryptedData} = await this.generateEncryptedDataDefault(objEncryptionKeysConfig);

		let nFirstDelimiterIndex = strEncryptedData.indexOf(Encryption.Encryption.ENCRYPTION_METADATA_DELIMITER, 0);

		assert(nFirstDelimiterIndex > 0, "No delimiter found.");

		//Remove only first metadata item: key index
		let strEncryptedDataWithoutKeyIndex = strEncryptedData.substring(nFirstDelimiterIndex + 1);

		try
		{
			let bufferDecryptedData = await encryptionInstance.decrypt(strEncryptedDataWithoutKeyIndex);
			let strDecryptedData = bufferDecryptedData.toString();

			this.logError(`[FAIL] testNoKeyIndex: Expected error to be thrown on decryption but got decrypted data: ${JSON.stringify(strDecryptedData)}.`);
		}
		catch(error)
		{
			console.log("[OK]", error.stack || error.message);
			this.logSuccess("[OK] testNoKeyIndex");
			return;
		}

		throw new Error("[FAIL] testNoKeyIndex: Expected an error to be thrown.");
	}

	/**
	 * Removes the cipher algorithm metadata from encrypted data.
	 * Expects that the decryption to be successful, using the default algorithm.
	 */
	async testNoCipherAlgorithm()
	{
		console.log("Running testNoCipherAlgorithm...");

		const strTestFeedData = this.constructor.FEED_DATA_DEFAULT;

		let objEncryptionKeysConfig = this.getContextClone();
		objEncryptionKeysConfig.keys[Encryption.Encryption.ENCRYPTION_KEY_INDEX_DEFAULT] = objEncryptionKeysConfig.keys[objEncryptionKeysConfig.active_index];

		let {encryptionInstance, strEncryptedData} = await this.generateEncryptedDataDefault(objEncryptionKeysConfig);

		let nFirstDelimiterIndex = strEncryptedData.indexOf(Encryption.Encryption.ENCRYPTION_METADATA_DELIMITER, 0);
		assert(nFirstDelimiterIndex > 0, "No delimiter found.");

		let nCipherAlgorithmDelimiterIndex = strEncryptedData.indexOf(Encryption.Encryption.ENCRYPTION_METADATA_DELIMITER, nFirstDelimiterIndex + 1);

		//Remove only second metadata item: cipher algorithm
		let strEncryptedDataWithoutCipherAlgorithmIndex = strEncryptedData.substring(0, nFirstDelimiterIndex + 1) + strEncryptedData.substring(nCipherAlgorithmDelimiterIndex + 1);

		let bufferDecryptedData = await encryptionInstance.decrypt(strEncryptedDataWithoutCipherAlgorithmIndex);
		let strDecryptedData = bufferDecryptedData.toString();

		assert(strTestFeedData === strDecryptedData, `Test failed: testNoCipherAlgorithm. Expected ${JSON.stringify(strTestFeedData)} and got ${JSON.stringify(strDecryptedData)}`);

		this.logSuccess("[OK] testNoCipherAlgorithm");
	}

	async testWrongKeyIndex()
	{
		console.log("Running testWrongKeyIndex...");

		let objEncryptionKeysConfig = this.getContextClone();
		objEncryptionKeysConfig.keys[Encryption.Encryption.ENCRYPTION_KEY_INDEX_DEFAULT] = objEncryptionKeysConfig.keys[objEncryptionKeysConfig.active_index];

		let {encryptionInstance, strEncryptedData} = await this.generateEncryptedDataDefault(objEncryptionKeysConfig);

		let nFirstDelimiterIndex = strEncryptedData.indexOf(Encryption.Encryption.ENCRYPTION_METADATA_DELIMITER, 0);
		assert(nFirstDelimiterIndex > 0, "No delimiter found.");

		//Rename key index to something wrong
		let strEncryptedDataWithWrongKeyIndex = "GIBBERISH-KEY" + Encryption.Encryption.ENCRYPTION_METADATA_DELIMITER + strEncryptedData.substring(nFirstDelimiterIndex + 1);

		try
		{
			let bufferDecryptedData = await encryptionInstance.decrypt(strEncryptedDataWithWrongKeyIndex);
			let strDecryptedData = bufferDecryptedData.toString();

			this.logError(`[FAIL] testWrongKeyIndex: Expected error to be thrown on decryption but got decrypted data: ${JSON.stringify(strDecryptedData)}.`);
		}
		catch(error)
		{
			console.log("[OK]", error.stack || error.message);
			this.logSuccess("[OK] testWrongKeyIndex");
			return;
		}

		throw new Error("[FAIL] testWrongKeyIndex: Expected an error to be thrown.");
	}

	async testWrongCipherAlgorithmInMetadata()
	{
		console.log("Running testWrongCipherAlgorithmInMetadata...");

		let objEncryptionKeysConfig = this.getContextClone();
		objEncryptionKeysConfig.keys[Encryption.Encryption.ENCRYPTION_KEY_INDEX_DEFAULT] = objEncryptionKeysConfig.keys[objEncryptionKeysConfig.active_index];

		let {encryptionInstance, strEncryptedData} = await this.generateEncryptedDataDefault(objEncryptionKeysConfig);

		let nFirstDelimiterIndex = strEncryptedData.indexOf(Encryption.Encryption.ENCRYPTION_METADATA_DELIMITER, 0);
		assert(nFirstDelimiterIndex > 0, "No delimiter found.");

		let nCipherAlgorithmDelimiterIndex = strEncryptedData.indexOf(Encryption.Encryption.ENCRYPTION_METADATA_DELIMITER, nFirstDelimiterIndex + 1);

		//Rename algorithm to something wrong
		let strEncryptedDataWithWrongCipherAlgorithm = strEncryptedData.substring(0, nFirstDelimiterIndex + 1) + "GIBBERISH-ALG" + Encryption.Encryption.ENCRYPTION_METADATA_DELIMITER + strEncryptedData.substring(nCipherAlgorithmDelimiterIndex + 1);

		try
		{
			let bufferDecryptedData = await encryptionInstance.decrypt(strEncryptedDataWithWrongCipherAlgorithm);
			let strDecryptedData = bufferDecryptedData.toString();

			this.logError(`[FAIL] testWrongCipherAlgorithmInMetadata: Expected error to be thrown on decryption but got decrypted data: ${JSON.stringify(strDecryptedData)}.`);
		}
		catch(error)
		{
			console.log("[OK]", error.stack || error.message);
			this.logSuccess("[OK] testWrongCipherAlgorithmInMetadata");
			return;
		}

		throw new Error("[FAIL] testWrongCipherAlgorithmInMetadata: Expected an error to be thrown.");
	}

	async testAlteredEncryptedData()
	{
		console.log("Running testAlteredEncryptedData...");

		let objEncryptionKeysConfig = this.getContextClone();
		objEncryptionKeysConfig.keys[Encryption.Encryption.ENCRYPTION_KEY_INDEX_DEFAULT] = objEncryptionKeysConfig.keys[objEncryptionKeysConfig.active_index];

		let {encryptionInstance, strEncryptedData} = await this.generateEncryptedDataDefault(objEncryptionKeysConfig);

		let nLastDelimiterIndex = -1;
		for(let i = 0; i < Encryption.Encryption.ENCRYPTION_METADATA_DELIMITER_COUNT; i++)
		{
			nLastDelimiterIndex = strEncryptedData.indexOf(Encryption.Encryption.ENCRYPTION_METADATA_DELIMITER, nLastDelimiterIndex + 1);
		}
		assert(nLastDelimiterIndex > 0, `No delimiter for position ${Encryption.Encryption.ENCRYPTION_METADATA_DELIMITER_COUNT}.`);

		//Remove all metadata
		let strEncryptedDataWithoutMetadata = strEncryptedData.substring(nLastDelimiterIndex + 1);

		let bufferFullEncryptedData = Buffer.from(strEncryptedDataWithoutMetadata, "base64");
		let strMetadata = strEncryptedData.substring(0, nLastDelimiterIndex + 1);
		let strIV = bufferFullEncryptedData.toString("binary", 0, Encryption.Encryption.ENCRYPTION_IV_LENGTH_BYTES_DEFAULT);

		let bufferNewEncryptedData = Buffer.concat(
			[
				Buffer.from(strIV, "binary"),
				Buffer.from("SOMETHING_TOTALLY_NEW_AND_GIBBERISH", "binary")
			]
		);

		let strNewEncryptedData = `${strMetadata}${bufferNewEncryptedData.toString("base64")}`;

		try
		{
			let bufferDecryptedData = await encryptionInstance.decrypt(strNewEncryptedData);
			let strDecryptedData = bufferDecryptedData.toString();

			this.logError(`[FAIL] testAlteredEncryptedData: Expected error to be thrown on decryption but got decrypted data: ${JSON.stringify(strDecryptedData)}.`);
		}
		catch(error)
		{
			console.log("[OK]", error.stack || error.message);
			this.logSuccess("[OK] testAlteredEncryptedData");
			return;
		}

		throw new Error("[FAIL] testAlteredEncryptedData: Expected an error to be thrown.");
	}

	async testCompromisedActiveKey()
	{
		console.log("Running testCompromisedActiveKey...");

		let objEncryptionKeysConfig = this.getContextClone();
		
		try
		{
			objEncryptionKeysConfig.keys[objEncryptionKeysConfig.active_index].compromised = true;
			await this.generateEncryptedDataDefault(objEncryptionKeysConfig);
		}
		catch(error)
		{
			if(error.message.toLowerCase().includes("active key") && error.message.includes("compromised"))
			{
				console.log("[OK]", error.stack || error.message);
			}
			else
			{
				throw error;
			}
		}

		try
		{
			
			let encryptionInstance = new Encryption.Encryption(
				this.getContextClone()
			);
			encryptionInstance._objKeyMap[objEncryptionKeysConfig.active_index].compromised = true;
			await encryptionInstance.encrypt(this.constructor.FEED_DATA_DEFAULT);
		}
		catch(error)
		{
			if(error.message.includes("compromised"))
			{
				console.log("[OK]", error.stack || error.message);
				this.logSuccess("[OK] testCompromisedActiveKey");
				return;
			}
			else
			{
				throw error;
			}
		}

		throw new Error("[FAIL] testCompromisedActiveKey: Expected an error to be thrown.");
	}

	async testCompromisedDefaultKey()
	{
		console.log("Running testCompromisedDefaultKey...");

		let objEncryptionKeysConfig = this.getContextClone();
		objEncryptionKeysConfig.keys[Encryption.Encryption.ENCRYPTION_KEY_INDEX_DEFAULT].compromised = true;

		try
		{
			let {encryptionInstance, strEncryptedData} = await this.generateEncryptedDataDefault(objEncryptionKeysConfig);
			let strEncryptedDataWithoutMetadata = this.removeAllMetadata(strEncryptedData);

			//This should throw an error
			await encryptionInstance.decrypt(strEncryptedDataWithoutMetadata);
		}
		catch(error)
		{
			if(error.message.toLowerCase().includes("default key") && error.message.includes("compromised"))
			{
				console.log("[OK]", error.stack || error.message);
				this.logSuccess("[OK] testCompromisedDefaultKey");
				return;
			}
			else
			{
				throw error;
			}
		}

		throw new Error("[FAIL] testCompromisedDefaultKey: Expected an error to be thrown.");
	}


	async testCompromisedKeyDecryption()
	{
		console.log("Running testCompromisedKeyDecryption...");

		const strNewKeyIndex = "k12345";
		const strNewKeyValue = await Encryption.Encryption.generateRandomString(32);
		const salt = await Encryption.Encryption.generateRandomString(32);

		try
		{
			let {strEncryptedData} = await this.generateEncryptedDataDefault();

			let objEncryptionKeysConfig = this.getContextClone();
			//Set previous active encryption key as compromised
			objEncryptionKeysConfig.keys[objEncryptionKeysConfig.active_index].compromised = true;

			//Create new encryption instance with another active key
			let encryptionInstanceForDecrypt = new Encryption.Encryption(
				{
					keys: Object.assign(
						objEncryptionKeysConfig.keys,
						{
							[strNewKeyIndex]: {
								aes_key: strNewKeyValue,
								salt: salt,
								created: "2018-12-31T23:51:55Z"
							}
						}
					),
					active_index: strNewKeyIndex
				}
			);

			//This should throw an error
			await encryptionInstanceForDecrypt.decrypt(strEncryptedData);
		}
		catch(error)
		{
			if(error.message.toLowerCase().includes("key") && error.message.includes("compromised") && error.message.includes("bAllowCompromisedKey"))
			{
				console.log("[OK]", error.stack || error.message);
				this.logSuccess("[OK] testCompromisedKeyDecryption");
				return;
			}
			else
			{
				throw error;
			}
		}

		throw new Error("[FAIL] testCompromisedKeyDecryption: Expected an error to be thrown.");
	}

	async testDecryptBadOptions()
	{
		console.log("Running testDecryptBadOptions...");
		let {strEncryptedData, encryptionInstance} = await this.generateEncryptedDataDefault();

		let arrBadOptions = [
			//bad types
			{skipHashValidation: "something"}, //Should be boolean
			{allowCompromisedKeys: null}, //Should be boolean
			{hasPayloadHash: 123}, //Should be boolean
			{hasPayloadIV: {}}, //Should be boolean
			{payloadBase64Encoded: "something"}, //Should be boolean
			{parseMetadata: "something"}, //Should be boolean
			{offsetBytesBeforePayloadParse: "something"}, //Should be Integer
			{hasPayloadIV: false, customIV: null}, //customIV should be non-empty string or Buffer
			{hasPayloadIV: false, customIV: ""} //customIV should be non-empty string or Buffer
		];

		for(let objBadOptions of arrBadOptions)
		{
			let bufferDecryptedData;
			try
			{
				bufferDecryptedData = await encryptionInstance.decrypt(strEncryptedData, objBadOptions);
			}
			catch(error)
			{
				console.log("[OK]", error.stack || error.message);
				continue;
			}

			throw new Error(`[FAIL] testDecryptBadOptions: Expected an error to be thrown for bad option ${JSON.stringify(objBadOptions)}, but got decrypted data "${bufferDecryptedData.toString()}"`);
		}

		this.logSuccess("[OK] testDecryptBadOptions");
	}

	async testEncryptRawAndCustomDecrypt()
	{
		console.log("Running testEncryptRaw...");

		const strTestFeedData = this.constructor.FEED_DATA_DEFAULT;
		const strKeyValue = await Encryption.Encryption.generateRandomString(32);
		const strSalt = await Encryption.Encryption.generateRandomString(16);
		const bufferIV = await Encryption.Encryption.generateRandomBytes(16);

		let {iv: bufferEncryptionIV, encryptedData: bufferEncryptedData} = await Encryption.Encryption.encryptRaw(strTestFeedData, strKeyValue, bufferIV);

		assert(Buffer.isBuffer(bufferEncryptedData) && bufferEncryptedData.byteLength > 0, "Returned encrypted buffer from encryptRaw must be a buffer with byteLength > 0.");
		assert(bufferEncryptionIV.equals(bufferIV), "Returned IV buffer from encryptRaw must be equal to given bufferIV.");

		let objEncryptionKeysConfig = this.getContextClone();

		objEncryptionKeysConfig.keys[Encryption.Encryption.ENCRYPTION_KEY_INDEX_DEFAULT] = {
			aes_key: strKeyValue,
			salt: strSalt,
			created: "2018-12-31T23:51:55Z"
		};

		let encryptionInstance = new Encryption.Encryption(objEncryptionKeysConfig);

		let objDecryptOptions = {
			hasPayloadHash: false,
			hasPayloadIV: false,
			customIV: bufferIV,
			payloadBase64Encoded: false,
			parseMetadata: false
		};

		let bufferDecryptedData = await encryptionInstance.decrypt(bufferEncryptedData, objDecryptOptions);

		let strDecryptedData = bufferDecryptedData.toString();
		assert(strTestFeedData === strDecryptedData, `Test failed: testEncryptRaw. Expected ${JSON.stringify(strTestFeedData)} and got ${JSON.stringify(strDecryptedData)}`);

		this.logSuccess("[OK] testEncryptRaw");
	}

	async testDecryptOffsetBytes()
	{
		console.log("Running testDecryptOffsetBytes...");

		const strTestFeedData = this.constructor.FEED_DATA_DEFAULT;
		const strKeyValue = await Encryption.Encryption.generateRandomString(32);
		const strSalt = await Encryption.Encryption.generateRandomString(16);
		const bufferIV = await Encryption.Encryption.generateRandomBytes(16);
		const bufferPrefix = Buffer.from("PREFIX_CUSTOM");

		let {iv: bufferEncryptionIV, encryptedData: bufferEncryptedData} = await Encryption.Encryption.encryptRaw(strTestFeedData, strKeyValue, bufferIV);

		assert(Buffer.isBuffer(bufferEncryptedData) && bufferEncryptedData.byteLength > 0, "Returned encrypted buffer from encryptRaw must be a buffer with byteLength > 0.");
		assert(bufferEncryptionIV.equals(bufferIV), "Returned IV buffer from encryptRaw must be equal to given bufferIV.");

		let objEncryptionKeysConfig = this.getContextClone();

		objEncryptionKeysConfig.keys[Encryption.Encryption.ENCRYPTION_KEY_INDEX_DEFAULT] = {
			aes_key: strKeyValue,
			salt: strSalt,
			created: "2018-12-31T23:51:55Z"
		};

		let encryptionInstance = new Encryption.Encryption(objEncryptionKeysConfig);

		let objDecryptOptions = {
			hasPayloadHash: false,
			hasPayloadIV: false,
			customIV: bufferIV,
			payloadBase64Encoded: true,
			offsetBytesBeforePayloadParse: bufferPrefix.byteLength,
			parseMetadata: false
		};
		bufferEncryptedData = Buffer.concat([
			bufferPrefix,
			bufferEncryptedData
		]);

		let bufferDecryptedData = await encryptionInstance.decrypt(bufferEncryptedData.toString("base64"), objDecryptOptions);

		let strDecryptedData = bufferDecryptedData.toString();
		assert(strTestFeedData === strDecryptedData, `Test failed: testDecryptOffsetBytes. Expected ${JSON.stringify(strTestFeedData)} and got ${JSON.stringify(strDecryptedData)}`);

		this.logSuccess("[OK] testDecryptOffsetBytes");
	}

	async testKeysPartition()
	{
		console.log("Running testKeysPartition...");

		const strPartitionName = "partition-1";
		const strKeyPrefix = strPartitionName + Encryption.Encryption.ENCRYPTION_PARTITION_DELIMITER;

		let objEncryptionKeysConfig = this.getContextClone();
		objEncryptionKeysConfig.keys_partition = strPartitionName;
		
		let {encryptionInstance, strEncryptedData} = await this.generateEncryptedDataDefault(objEncryptionKeysConfig);
		console.log(`testKeysPartition: encryptedData -> ${strEncryptedData}`);
		
		assert(encryptionInstance.hasPartition(), "When partition is set, hasPartition() must return true.");
		assert(strEncryptedData.startsWith(strKeyPrefix), `Encrypted data does not contain partition prefix: ${strKeyPrefix}`);

		let strExpectedActiveKeyWithPartition = strKeyPrefix + encryptionInstance.activeKeyIndex;
		assert(
			encryptionInstance._getKeyNameWithPartition(encryptionInstance.activeKeyIndex) === strExpectedActiveKeyWithPartition,
			`Wrong format for _getKeyNameWithPartition. Expected ${strExpectedActiveKeyWithPartition}, but got ${encryptionInstance._getKeyNameWithPartition(encryptionInstance.activeKeyIndex)}`
		);

		let bufferDecryptedData = await encryptionInstance.decrypt(strEncryptedData);

		let strDecryptedData = bufferDecryptedData.toString();

		assert(this.constructor.FEED_DATA_DEFAULT === strDecryptedData, `Test failed: testKeysPartition. Expected ${JSON.stringify(this.constructor.FEED_DATA_DEFAULT)} and got ${JSON.stringify(strDecryptedData)}`);

		encryptionInstance.partition = null;

		assert(
			encryptionInstance.hasPartition() === false,
			"After partition was set to null, hasPartition() should have been false."
		);

		strExpectedActiveKeyWithPartition = encryptionInstance.activeKeyIndex;
		assert(
			encryptionInstance._getKeyNameWithPartition(encryptionInstance.activeKeyIndex) === strExpectedActiveKeyWithPartition,
			`Wrong format for _getKeyNameWithPartition. Expected ${strExpectedActiveKeyWithPartition}, but got ${encryptionInstance._getKeyNameWithPartition(encryptionInstance.activeKeyIndex)}`
		);

		this.logSuccess("[OK] testKeysPartition");
	}

	async testMultipleRandomEncryption()
	{
		const nRunsCount = this.context.randomEncryptionRuns || this.constructor.RANDOM_ENCRYPTION_RUNS_DEFAULT;
		console.log(`Running testMultipleRandomEncryption with ${nRunsCount} runs with test random data length of max ${this.constructor.RANDOM_ENCRYPTION_FEED_DATA_LENGTH_MAX} characters...`);
		const nStartMs = (new Date()).getTime();

		const fnTestRun = async(nRunID) => {
			const nTestFeedDataLength = parseInt(Math.random() * this.constructor.RANDOM_ENCRYPTION_FEED_DATA_LENGTH_MAX, 10);
			const strTestFeedData = await Encryption.Encryption.generateRandomString(nTestFeedDataLength, "ascii");
			assert(strTestFeedData.length === nTestFeedDataLength, `Generated test data for run #${nRunID} expected to be ${nTestFeedDataLength}, but got a string of length ${strTestFeedData.length}.`);

			let objEncryptionKeysConfig = this.getContextClone();
			objEncryptionKeysConfig.keys[objEncryptionKeysConfig.active_index].aes_key = await Encryption.Encryption.generateRandomString(32, "ascii");
			objEncryptionKeysConfig.keys[objEncryptionKeysConfig.active_index].salt = await Encryption.Encryption.generateRandomString(32, "ascii");

			let {encryptionInstance, strEncryptedData} = await this.generateEncryptedDataDefault(objEncryptionKeysConfig, strTestFeedData);

			let bufferDecryptedData = await encryptionInstance.decrypt(strEncryptedData);

			let strDecryptedData = bufferDecryptedData.toString();

			assert(strTestFeedData === strDecryptedData, `Test failed: testMultipleRandomEncryption on run #${nRunID} of ${nRunsCount}. Expected ${JSON.stringify(strTestFeedData)} and got ${JSON.stringify(strDecryptedData)}`);
		};

		let arrPromises = [];
		for(let i = 0; i < nRunsCount; i++)
		{
			arrPromises.push(fnTestRun(i));
		}

		await Promise.all(arrPromises);

		const nFinishMs = (new Date()).getTime();
		this.logSuccess(`[OK] testMultipleRandomEncryption with ${nRunsCount} runs in ${nFinishMs - nStartMs} ms`);
	}

	/**
	 *
	 * @param {string} strEncryptedData
	 *
	 * @returns {string}
	 */
	removeAllMetadata(strEncryptedData)
	{
		let nLastDelimiterIndex = -1;
		for(let i = 0; i < Encryption.Encryption.ENCRYPTION_METADATA_DELIMITER_COUNT; i++)
		{
			nLastDelimiterIndex = strEncryptedData.indexOf(Encryption.Encryption.ENCRYPTION_METADATA_DELIMITER, nLastDelimiterIndex + 1);
		}
		assert(nLastDelimiterIndex > 0, `No delimiter for position ${Encryption.Encryption.ENCRYPTION_METADATA_DELIMITER_COUNT}.`);

		//Remove all metadata
		return strEncryptedData.substring(nLastDelimiterIndex + 1);
	}

	/**
	 * @param {Object} [objEncryptionKeysConfig=null]
	 * @param {string} [strTestFeedData=EncryptionTest.FEED_DATA_DEFAULT]
	 * 
	 * @returns {{encryptionInstance: Encryption.Encryption, strEncryptionData: string}}
	 */
	async generateEncryptedDataDefault(objEncryptionKeysConfig = null, strTestFeedData = EncryptionTest.FEED_DATA_DEFAULT)
	{
		objEncryptionKeysConfig = objEncryptionKeysConfig || this.context;

		let encryptionInstance = new Encryption.Encryption(
			objEncryptionKeysConfig,
			Encryption.Encryption.CIPHER_ALGORITHM_AES_CBC
		);

		let strEncryptedData = await encryptionInstance.encrypt(strTestFeedData);
		this.validateEncryptedDataFormat(strEncryptedData);

		return {
			encryptionInstance,
			strEncryptedData
		};
	}

	/**
	 * Format: "KEY_INDEX_NAME|ALGORITHM_NAME_LOWER_CASE|" + base64(iv + encrypted_data)
	 *
	 * @param {string} strEncryptedData
	 */
	validateEncryptedDataFormat(strEncryptedData)
	{
		assert(
			typeof strEncryptedData === "string" && strEncryptedData.trim().length > 0,
			`Invalid type for encrypted data. Expected "string", but got "${typeof strEncryptedData}".`
		);

		assert(strEncryptedData.includes(Encryption.Encryption.ENCRYPTION_METADATA_DELIMITER), `Encryption data does not contain any metadata delimiter character '${Encryption.Encryption.ENCRYPTION_METADATA_DELIMITER}'.`);

		const nMetadataDelimitersFound = strEncryptedData.split(Encryption.Encryption.ENCRYPTION_METADATA_DELIMITER).length - 1;
		assert(
			nMetadataDelimitersFound === Encryption.Encryption.ENCRYPTION_METADATA_DELIMITER_COUNT,
			`Expected that encrypted data to contain a count of metadata delimiter character equal to ${Encryption.Encryption.ENCRYPTION_METADATA_DELIMITER_COUNT}, but ${nMetadataDelimitersFound} found.`
		);
	}

	static get FEED_DATA_DEFAULT() { return "something to test123!"; }
	static get RANDOM_ENCRYPTION_RUNS_DEFAULT() { return 2000; }
	static get RANDOM_ENCRYPTION_FEED_DATA_LENGTH_MAX() { return 2500; }
};
