"use strict";

const ISCSINameValidatorException = require("./ISCSINameValidatorException");
const DomainNameValidator = require("../DomainName/DomainNameValidator");
const BaseValidator = require("../Base/BaseValidator");

module.exports =
class ISCSINameValidator extends BaseValidator
{
	constructor()
	{
		throw new Error("ISCSINameValidator is a static class and cannot be instantiated.");
	}

	/**
	* Returns true if strISCSIName is valid, throws otherwise (does not return false)
	*
	* @param {string} strISCSIName
	*
	* @returns {boolean}
	*
	* @throws {ISCSINameValidatorException | DomainNameValidatorException}
	* ISCSINameValidatorException codes: INVALID_TYPE_DESIGNATOR, NO_IQN_MANDATORY_PART, WRONG_DATE_FORMAT
	**/
	static assertValid(strISCSIName)
	{
		if(typeof strISCSIName !== "string" || strISCSIName.trim().length === 0)
		{
			throw new ISCSINameValidatorException(`Invalid value type "${typeof strISCSIName}" for iSCSI name. Only accepts non-empty "string".`);
		}

		let position = strISCSIName.indexOf(".");
		let arrTypeAndDetails = [strISCSIName.slice(0, position), strISCSIName.slice(position + 1)];

		if (arrTypeAndDetails[0] === "iqn")
		{
			return ISCSINameValidator.checkIQNFormat(arrTypeAndDetails[1]);
		}
		else if (arrTypeAndDetails[0] === "eui")
		{
			return ISCSINameValidator.checkEUIFormat(arrTypeAndDetails[1]);
		}
		else
		{
			return new ISCSINameValidatorException("Invalid identifier type designator", ISCSINameValidatorException.INVALID_TYPE_DESIGNATOR);
		}
	}

	/** RFC 3721 ; 1.1
	* Utility function.
	* Returns true if strNameDetails is a valid iSCSI name, throws otherwise
	* Usage example : ISCSINameValidator.checkIQNFormat("1997-12.com.example:example-name")
	*
	* @param {string} strNameDetails
	* @returns {boolean}
	* @throws {ISCSINameValidatorException | DomainNameValidatorException}
	* ISCSINameValidatorException codes: NO_IQN_MANDATORY_PART, WRONG_DATE_FORMAT
	*/
	static checkIQNFormat(strNameDetails)
	{
		if(typeof strNameDetails !== "string" || strNameDetails.trim().length === 0)
		{
			throw new ISCSINameValidatorException(`Invalid value type "${typeof strNameDetails}" for iSCSI IQN format. Only accepts non-empty "string".`);
		}

		let position = strNameDetails.indexOf(":");
		if (position === -1)
		{
			throw new ISCSINameValidatorException("Date and issuer not provided", ISCSINameValidatorException.NO_IQN_MANDATORY_PART);
		}

		let arrNameDetails = [strNameDetails.slice(0, position), strNameDetails.slice(position + 1)];

		let strDateAndIssuer = arrNameDetails[0];
		let boolDateAndIssuerValid = this.checkDateAndIssuer(strDateAndIssuer);

		let strIssuerDefinedName = arrNameDetails[1];
		// If colon is present, issuer-defined string cannot be empty
		if (strIssuerDefinedName.indexOf(":") > -1 && strIssuerDefinedName.indexOf(":") == (strIssuerDefinedName.length - 1))
		{
			throw new ISCSINameValidatorException("Issuer-defined string is empty", ISCSINameValidatorException.NO_IQN_MANDATORY_PART);
		}

		return boolDateAndIssuerValid;
	}

	/** Utility function
	* Returns true if strDateAndIssuer is valid (issuer valid, 1<=month<=12)
	* Usage example : ISCSIValidator.checkDateAndIssuer("1997-12.com.example")
	*
	* @param {string} $strDateAndIssuer
	* @returns {boolean}
	* @throws {ISCSINameValidatorException | DomainNameValidatorException}
	* ISCSINameValidatorException codes: WRONG_DATE_FORMAT
	**/
	static checkDateAndIssuer(strDateAndIssuer)
	{
		if(typeof strDateAndIssuer !== "string" || strDateAndIssuer.trim().length === 0)
		{
			throw new ISCSINameValidatorException(`Invalid value type "${typeof strDateAndIssuer}" for iSCSI date and issuer. Only accepts non-empty "string".`);
		}

		let position = strDateAndIssuer.indexOf(".");

		if (position === -1)
		{
			throw new ISCSINameValidatorException("Date or issuer missing", ISCSINameValidatorException.NO_IQN_MANDATORY_PART);
		}

		let arrDateAndIssuer = [strDateAndIssuer.slice(0, position), strDateAndIssuer.slice(position + 1)];

		let strDate = arrDateAndIssuer[0];
		let arrDate = strDate.split("-");
		if (arrDate.length !== 2)
		{
			throw new ISCSINameValidatorException("Invalid date format", ISCSINameValidatorException.WRONG_DATE_FORMAT);
		}

		if (Number(arrDate[1]) < 1 || Number(arrDate[1]) > 12)
		{
			throw new ISCSINameValidatorException("Invalid date format", ISCSINameValidatorException.WRONG_DATE_FORMAT);
		}

		let strReverseIssuerDomain = arrDateAndIssuer[1];
		let strIssuerDomain = "www." + strReverseIssuerDomain.split(".").reverse().join(".");

		// The domain must be valid and must have a TLD label
		DomainNameValidator.assertValid(strIssuerDomain, false);

		return true;
	}

	/** RFC 3721 ; 1.2
	* Utility function.
	* Returns true if strNameDetails is a valid iSCSI name, throws otherwise
	* Usage example : ISCSINameValidator.checkEUIFormat("02004567A425678D")
	*
	* @param {String} strNameDetails
	* @returns {boolean}
	* @throws {ISCSINameValidatorException}
	* ISCSINameValidatorException codes: NO_HEXADECIMAL, NOT_64_BIT
	*/
	static checkEUIFormat(strNameDetails)
	{
		let objHexRegex = /^[0-9A-Fa-f]*$/;
		if (!objHexRegex.test(strNameDetails))
		{
			throw new ISCSINameValidatorException("Non-hexadecimal character found", ISCSINameValidatorException.NOT_HEXADECIMAL);
		}

		if (strNameDetails.length != 16)
		{
			throw new ISCSINameValidatorException("Not a 64-bit (16 hex characters) value", ISCSINameValidatorException.NOT_64_BIT)
		}

		return true;
	}

	/**
	* Utility function.
	* Tests if the iSCSI names in the arrays are correctly validated
	* Should only output "OK" to the console if the validator is working properly
	*/
	static test()
	{
		let arrValidISCSINames = [
			"iqn.2001-04.com.example:storage:strNameDetails",
			"iqn.2001-04.com.example:strNameDetails",
			"iqn.2011-11.com.example:storage:strNameDetails",
			"eui.020B04567A45678D"
		];

		let arrInvalidISCSINames = [
			"isn.2001-04.com.example:storage:strNameDetails",
			"iqn.2001-04.com.example:storage:",
			"iqn.2001-22.com.example:storage:strNameDetails",
			"eui.11111111111111111111",
			"eui.125R123141212415",
			"eui.000000000000000G"
		];

		for (let strISCSIName of arrValidISCSINames)
		{
			if(ISCSINameValidator.isValid(strISCSIName))
			{
				console.log("OK");
			}
			else
			{
				console.log("Valid iSCSIName " + strISCSIName + " erroneously detected as invalid");
			}
		}

		for (let strISCSIName of arrInvalidISCSINames)
		{
			if(!ISCSINameValidator.isValid(strISCSIName))
			{
				console.log("OK");
			}
			else
			{
				console.log("Invalid iSCSIName " + strISCSIName + " erroneously detected as valid");
			}
		}
	}
};
