"use strict";

const TimestampValidatorException = require("./TimestampValidatorException");
const DomainNameValidator = require("../DomainName/DomainNameValidator");
const BaseValidator = require("../Base/BaseValidator");

module.exports =
class TimestampValidator extends BaseValidator
{
	constructor()
	{
		throw new Error("TimestampValidator is a static class and cannot be instantiated.");
	}

	/**
	* Returns true if strTimestamp is valid, throws otherwise (does not return false)
	*
	* @param {string} strTimestamp
	* @returns {boolean}
	* @throws {TimestampValidatorException}
	* TimestampValidatorException codes: DATE_MUST_BE_STRING, DATE_MUST_BE_ISO8601_COMPLIANT, INVALID_DATE
	**/
	static assertValid(strTimestamp)
	{
		if (typeof strTimestamp !== "string" || strTimestamp.trim().length === 0)
		{
			throw new TimestampValidatorException("The date must be a non-empty string.", TimestampValidatorException.DATE_MUST_BE_STRING);
		}

		let objDateRegex = /^(\d{4})-(\d{2})-(\d{2})T(\d{2}):(\d{2}):(\d{2})Z$/;
		let arrDateComponents = objDateRegex.exec(strTimestamp);

		if (arrDateComponents === null)
		{
			throw new TimestampValidatorException("The date is not ISO-8601 compliant and in the expected format.", TimestampValidatorException.DATE_MUST_BE_ISO8601_COMPLIANT);
		}

		let objDate = new Date(arrDateComponents[1], arrDateComponents[2] - 1, arrDateComponents[3]);
		if (objDate.getFullYear() != arrDateComponents[1] || objDate.getMonth() != arrDateComponents[2] - 1 || objDate.getDate() != arrDateComponents[3])
		{
			throw new TimestampValidatorException("Invalid date in timestamp.", TimestampValidatorException.INVALID_DATE);
		}

		if (
			arrDateComponents[4] < 0 || arrDateComponents[4] >= 24
			|| arrDateComponents[5] < 0 || arrDateComponents[5] >= 60
			|| arrDateComponents[6] < 0 || arrDateComponents[6] >= 60
		)
		{
			throw new TimestampValidatorException("Invalid time in timestamp.", TimestampValidatorException.INVALID_TIME);
		}

		return true;
	}

	/**
	* Utility function.
	* Tests if the timestamps in the arrays are correctly validated
	* Should only output "OK" to the console if the validator is working properly
	*/
	static test()
	{
		let arrValidTimestamps = [
			"2013-02-28T23:59:07Z",
			"2010-01-21T10:10:10Z",
			"2013-10-11T01:00:00Z"
		];

		let arrInvalidTimestamps = [
			"2013-02-30T23:59:07Z",
			"2013-02-28T23:59:07C",
			"2013-02-28M23:59:07Z",
			"2013-02--28T23:59:07Z",
			"2013-22-28T23:59:07Z",
			"2011-01-35T23:59:07Z",
			"ZZZZZZZZZZZZZZZZZZZZ",
			"2000-00-00T10:10:10Z",
			"2000-01-01T25:30:30Z"
		];

		for (let strTimestamp of arrValidTimestamps)
		{
			if(TimestampValidator.isValid(strTimestamp))
			{
				console.log("OK");
			}
			else
			{
				console.log("Valid timestamp " + strTimestamp + " erroneously detected as invalid");
			}
		}

		for (let strTimestamp of arrInvalidTimestamps)
		{
			if(!TimestampValidator.isValid(strTimestamp))
			{
				console.log("OK");
			}
			else
			{
				console.log("Invalid timestamp " + strTimestamp + " erroneously detected as valid");
			}
		}
	}
};
