"use strict";

const URLValidatorException = require("./URLValidatorException");
const DomainNameValidator = require("../DomainName/DomainNameValidator");
const BaseValidator = require("../Base/BaseValidator");

module.exports = 
class URLValidator extends BaseValidator
{
	constructor()
	{
		throw new Error("URLValidator is a static class and cannot be instantiated.");
	}

	static isValid(strURL, strValidationType, bAllowNonIANATLD)
	{
		try
		{
			if (strValidationType === this.VALIDATION_TYPE_URL) 
			{
				return this.assertValid(strURL, bAllowNonIANATLD);
			}
			else
			{
				throw new Error(`Validation Type ${strValidationType} is not supported.`);
			}
		}
		catch(error)
		{
			return false;
		}	
	}

	/**
	* Validates using RFC 3986.
	* Only allows for HTTP and HTTPS URI schemes.
	*
	*     foo://example.com:8042/over/there?name=ferret#nose
	*     \_/   \______________/\_________/ \_________/ \__/
	*      |           |            |            |        |
	*   scheme     authority       path        query   fragment
	*      |   _____________________|__
	*     / \ /                        \
	*     urn:example:animal:ferret:nose
	*
	* @param {string} strURL 
	* @returns {boolean}
	*
	* @throws {DomainNameValidatorException | URLValidatorException}
	*/
	static assertValid(strURL, bAllowNonIANATLD, bAllowHash, bAllowIPs)
	{
		if(typeof strURL !== "string" || strURL.trim().length === 0)
		{
			throw new URLValidatorException(`Invalid value type "${typeof strURL}" for timestamp. Only accepts non-empty "string".`);
		}

		//It does not matter for validation, and simplifies lookups.
		strURL = strURL.toLowerCase();
		
		if(strURL.length > this.URL_MAX_LENGTH)
		{
			throw new URLValidatorException(`URL cannot be longer than ${this.URL_MAX_LENGTH} characters. Found ${strURL.length} in ${JSON.stringify(strURL)}`, URLValidatorException.URL_LONGER_THAN_MAXIMUM_ALLOWED);
		}
		
		let arrAllowedURLSchemes = this.URL_ALLOWED_SCHEMES;
		
		let arrSchemeAndRest = strURL.split(":");
		let strScheme = arrSchemeAndRest[0].toLowerCase();
		let strRemainingURL;
		
		if(arrSchemeAndRest.length >= 2)
		{
			for(let i = 1; i < arrSchemeAndRest.length; i++)
			{
				if(i == 1)
					strRemainingURL = arrSchemeAndRest[i];
				else
					strRemainingURL = strRemainingURL + ":" + arrSchemeAndRest[i];
			}
		}
		else
		{
			throw new URLValidatorException(`Missing scheme separator ':' for URL ${JSON.stringify(strURL)}.`, URLValidatorException.MISSING_SCHEME_SEPARATOR);
		}
			
		if(arrAllowedURLSchemes.indexOf(strScheme) == -1)
		{
			throw new URLValidatorException(`URL scheme ${JSON.stringify(strScheme)} for URL ${JSON.stringify(strURL)} not supported. Allowed ones are: ${JSON.stringify(arrAllowedURLSchemes)}.`, URLValidatorException.SCHEME_NOT_ALLOWED);
		}
		
		if(strRemainingURL.substr(0, 2) != "//")
		{
			throw new URLValidatorException(`Missing authority path-abempty (missing '//' after URI scheme and ':') in ${JSON.stringify(strURL)}.`, URLValidatorException.MISSING_AUTHORITY_PATH_ABEMPTY);
		}

		strRemainingURL = strRemainingURL.substr(2);

		
		let arrHierarchyPathAndRest = strRemainingURL.split("/");
		let strUserPasswordAndDomainName = arrHierarchyPathAndRest[0];
		let strPathQueryAndFragment;
		
		if(arrHierarchyPathAndRest.length >= 2)
		{
			for(let i = 1; i < arrHierarchyPathAndRest.length; i++)
			{
				if(i == 1)
				{
					strPathQueryAndFragment=arrHierarchyPathAndRest[i];
				}
				else
				{
					strPathQueryAndFragment=strPathQueryAndFragment+"/"+arrHierarchyPathAndRest[i];
				}
			}
		}
		else
		{
			strPathQueryAndFragment="";
		}
			
			
		/**
		* RFC 3986. 
		* 3.2.1.  User Information.
		* Use of the format "user:password" in the userinfo field is
		* deprecated.  Applications should not render as clear text any data
		* after the first colon (":") character found within a userinfo
		* subcomponent unless the data after the colon is the empty string
		* (indicating no password).  Applications may choose to ignore or
		* reject such data when it is received as part of a reference and
		* should reject the storage of such data in unencrypted form.  The
		* passing of authentication information in clear text has proven to be
		* a security risk in almost every case where it has been used.
		*/
		
		let arrUserInfoAndHostPortParts = strUserPasswordAndDomainName.split("@");
		let arrUserAndPassword = [];
		
		let myArray;
		
		if(arrUserInfoAndHostPortParts.length >= 2)
		{
			arrUserAndPassword = arrUserInfoAndHostPortParts.shift();
			arrUserAndPassword = arrUserAndPassword.split(":");
			if(arrUserAndPassword.length === 2)
			{
				if(arrUserAndPassword[1] != "")
				{
					throw new URLValidatorException("RFC 3986 deprecates storing password information on URLs as it is sure to be a security issue.", URLValidatorException.USER_INFO_PASSWORD_NOT_ALLOWED);
				}
			}

			let strUserName=arrUserAndPassword[0];
			
			/**
			* RFC 3986.
			* 3.2.1.  User Information.
			* userinfo = *( unreserved / pct-encoded / sub-delims / ":" ).
			* 
			* pct-encoded   = "%" HEXDIG HEXDIG
			* unreserved    = ALPHA / DIGIT / "-" / "." / "_" / "~"
			* sub-delims    = "!" / "$" / "&" / "'" / "(" / ")" / "*" / "+" / "," / ";" / "="
			*/
			
			let myRegex = /%[0-9a-f]{2,2}/gi;
			strUserName = strUserName.replace(myRegex, "");
			
			// myArray=["!", "\$", "&", "'", "(", ")", "*", "+", ",", ";", "="];
			myArray=[/!/g, /\$/g, /&/g, /'/g, /\(/g, /\)/g, /\*/g, /\+/g, /,/g, /;/g, /=/g];
			
			for(let key in myArray)
			{
				strUserName = strUserName.replace(myArray[key], "");
			}
			
			myRegex = /[0-9a-z]/gi;
			strUserName = strUserName.replace(myRegex, "");
			
			if(strUserName.length)
			{
				throw new URLValidatorException(`Found invalid characters in username: ${strUserName} for URL ${strURL}.`, URLValidatorException.USERNAME_INVALID);
			}
		}
		
		let arrDomainNameAndPort = arrUserInfoAndHostPortParts[0].split(":");
		let strPort;
		
		if(arrDomainNameAndPort.length >= 2)
		{
			for(let i = 1; i < arrDomainNameAndPort.length; i++)
			{
				if(i == 1)
				{
					strPort = arrDomainNameAndPort[i];
				}
				else 
				{
					strPort = strPort+":"+arrDomainNameAndPort[i];
				}
			}
			
			if(isNaN(strPort) || strPort.substr(0, 1) == "0" || Number(strPort) <= 0 || Number(strPort) > 65535)
			{
				throw new URLValidatorException(`Invalid TCP port: ${strPort} for URL ${strURL}.`, URLValidatorException.INVALID_PORT);
			}
		}
		
		let strDomainName = arrDomainNameAndPort[0];
		DomainNameValidator.assertValid(strDomainName, bAllowNonIANATLD, bAllowIPs);

		let arrPathAndRest = strPathQueryAndFragment.split("?");
		
		let strPath = arrPathAndRest[0];
		let strQueryAndFragment;
		
		if(arrPathAndRest.length >= 2)
		{
			for(let i = 1; i < arrPathAndRest.length; i++)
			{
				if(i == 1)
				{
					strQueryAndFragment = arrPathAndRest[i];
				}
				else 
				{
					strQueryAndFragment = strQueryAndFragment + "?" + arrPathAndRest[i];
				}
			}
		}
		else
		{
			strQueryAndFragment = "";
		}
		
		
		/**
		* Path types:
		*
		* path-abempty: begins with "/" or is empty
		* path-absolute: begins with "/" but not "//"
		* path-noscheme: begins with a non-colon segment
		* path-rootless: begins with a segment
		* path-empty: zero characters
		*
		* path-abempty  = *( "/" segment )
		* path-absolute = "/" [ segment-nz *( "/" segment ) ]
		* path-noscheme = segment-nz-nc *( "/" segment )
		* path-rootless = segment-nz *( "/" segment )
		* path-empty    = 0<pchar>
		*
		* segment       = *pchar
		* segment-nz    = 1*pchar
		* segment-nz-nc = 1*( unreserved / pct-encoded / sub-delims / "@" )
		* 			; non-zero-length segment without any colon ":"
		* 
		* pchar         = unreserved / pct-encoded / sub-delims / ":" / "@"
		*/
		
		
		myArray = [/:/g, /@/g, /\//g];
		for(let key in myArray)
		{
			strPath = strPath.replace(myArray[key], "");
		}
		
		strPath = URLValidator.replace_pchar(strPath);
		
		if(strPath.length)
		{
			throw new URLValidatorException(`Found invalid characters in URL path ${strPath}`, URLValidatorException.PATH_INVALID);
		}
			
		let arrQueryAndFragment = strQueryAndFragment.split("#");
		let strQuery = arrQueryAndFragment[0];
		let strFragment;
		
		if(arrQueryAndFragment.length >= 2)
		{
			for(let i = 1; i < arrQueryAndFragment.length; i++)
			{
				if(i == 1)
				{
					strFragment = arrQueryAndFragment[i];
				}
				else
				{
					strFragment = strFragment + "#" + arrQueryAndFragment[i];
				}
			}
		}
		else
		{
			strFragment = "";
		}
		
		/**
		*  query       = *( pchar / "/" / "?" )
		*/
		
		strQuery = strQuery.replace(/\//g, "");
		strQuery = strQuery.replace(/\?/g, "");
		
		strQuery = URLValidator.replace_pchar(strQuery);
		
		if(strQuery.length)
		{
			throw new URLValidatorException(`Found invalid characters in URL query: ${strQuery}.`, URLValidatorException.QUERY_INVALID);
		}
		
		/**
		*  fragment    = *( pchar / "/" / "?" )
		*/
		
		strFragment = strFragment.replace(/\//g, "");
		strFragment = strFragment.replace(/\?/g, "");
		
		strFragment = URLValidator.replace_pchar(strFragment);
		
		if(strFragment.length)
		{
			throw new URLValidatorException(`Found invalid characters in URL fragment: ${strFragment}`, URLValidatorException.FRAGMENT_INVALID);
		}
		
		return true;
	}


	static replace_pchar(strURL)
	{
		/**
		* pchar         = unreserved / pct-encoded / sub-delims / ":" / "@"
		* unreserved    = ALPHA / DIGIT / "-" / "." / "_" / "~"
		* sub-delims    = "!" / "$" / "&" / "'" / "(" / ")"
		* 				/ "*" / "+" / "," / ";" / "="
		* pct-encoded   = "%" HEXDIG HEXDIG
		*/
		
		let myRegex = /[0-9a-z\.~_-]/gi;
		strURL = strURL.replace(myRegex, "");

		let myArray = [/!/g, /\$/g, /&/g, /'/g, /\(/g, /\)/g, /\*/g, /\+/g, /,/g, /;/g, /=/g];
		for(let key in myArray)
		{
			strURL = strURL.replace(myArray[key], "");
		}
			
		myRegex = /%[0-9a-f]{2,2}/gi;
		strURL = strURL.replace(myRegex, "");
		
		
		strURL = strURL.replace(/:/g, "");
		strURL = strURL.replace(/@/g, "");
		
		return strURL;
	}

	static get VALIDATION_TYPE_URL() { return "url"; }

	static get URL_MAX_LENGTH() { return 4096; }
	static get URL_ALLOWED_SCHEMES() { return ["http", "https", "ws"]; }
};
