"use strict";

const CacheEngine = require("./Engines/CacheEngine");
const RedisCacheEngine = require("./Engines/RedisCacheEngine");
const InMemoryCacheEngine = require("./Engines/InMemoryCacheEngine");

module.exports =
class CacheFactory
{
    constructor()
    {
        throw new Error("CacheFactory is a static class.");
    }

    /**
     * List with all supported Cache Engines.
     * 
     *  @type {Array}
     */
    static get supportedEngines() { return [this.ENGINE_REDIS, this.ENGINE_IN_MEMORY]; }

    /**
     * Default engine in case the engine type is not set in config.
     * 
     * @type {String}
     */
    static get defaultEngine() { return this.ENGINE_REDIS; }

    /**
     * Creates a new CacheEngine based on the given cache engine type.
     * 
     * @param {String} strCacheEngineType 
     * @param {Object} objCacheEngineConfig 
     * 
     * @return {CacheEngine}
     */
    static makeEngine(strCacheEngineType = CacheFactory.defaultEngine, objCacheEngineConfig = {})
    {
        if(this.supportedEngines.indexOf(strCacheEngineType) < 0)
        {
            throw new Error(`Cache engine of type "${strCacheEngineType}" is not supported. Supported types are: ${JSON.stringify(this.supportedEngines)}.`);
        }

        switch(strCacheEngineType)
        {
            case this.ENGINE_REDIS:
                return new RedisCacheEngine(objCacheEngineConfig);
            case this.ENGINE_IN_MEMORY:
                return new InMemoryCacheEngine(objCacheEngineConfig);
            default:
                throw new Error(`Unhandled cache engine type "${strCacheEngineType}".`);
        }
    }

    static get ENGINE_REDIS() { return "redis"; }
    static get ENGINE_MEMCACHE() { return "memcache"; }
    static get ENGINE_COUCHBASE() { return "couchbase"; }
    static get ENGINE_IN_MEMORY() { return "in_memory"; }
};
