"use strict";

const CacheEngine = require("./CacheEngine");
const CacheResult = require("../CacheResult");
const redis = require("redis");

module.exports =
class RedisCacheEngine extends CacheEngine
{
	constructor(objConfig)
	{
		super(objConfig);
		//it is initiated on getter for client.
		this._redisClient = null;
	}

	/**
	 * @type {RedisClient}
	 */
	get client() 
	{
		if(
			typeof this._redisClient == "undefined"
			|| this._redisClient == null
		)
		{
			this._redisClient = redis.createClient(this.config);

			this.constructor._setClientEventHandlers(this._redisClient);
			this.constructor._promisifyClient(this._redisClient);
		}

		return this._redisClient;
	}

	/**
	 * Fetch the value of a cache key.
	 * 
	 * @returns {CacheResult}
	 */
	async fetch(strCacheKey)
	{
		let result = await this.client.getAsync(this._sandboxTheCacheKey(strCacheKey));
		if(result == null)
		{
			return CacheResult.staticMissSingleton;
		}

		return new CacheResult(true, this.constructor.deserialize(result));
	}
	

	/**
	 * Sets the value of a cache key.
	 * https://redis.io/commands/set
	 * https://www.npmjs.com/package/redis#sending-commands
	 * 
	 * @param {String} strCacheKey
	 * @param {*} mxValue
	 * @param {Integer} nValidSeconds = null
	 */
	async store(strCacheKey, mxValue, nValidSeconds = null)
	{
		if(nValidSeconds != null)
		{
			return this.client.setAsync(
				this._sandboxTheCacheKey(strCacheKey), 
				this.constructor.serialize(mxValue), 
				'EX', 
				nValidSeconds
			);
		}
		else
		{
			return this.client.setAsync(
				this._sandboxTheCacheKey(strCacheKey), 
				this.constructor.serialize(mxValue)
			);
		}
	}

	/**
	 * Expires a key.
	 * 
	 * @param {String} strCacheKey 
	 */
	async expire(strCacheKey)
	{
		return this.client.delAsync(this._sandboxTheCacheKey(strCacheKey));
	}
	
	/**
	 * Clear all cache data.
	 */
	async destroy()
	{
		return this.client.flushdbAsync();
	}

	/**
	 * 
	 * @param {function} fnErrorHandler 
	 */
	setErrorHandler(fnErrorHandler)
	{
		if(typeof fnErrorHandler !== "function")
		{
			throw new TypeError(`Invalid type of error handler function "${typeof fnErrorHandler}".`);
		}

		this.client.on("error", fnErrorHandler);
	}

	/**
	 * Adds Async methods to redis client functions with callbacks.
	 * The new functions contain the suffix Async.
	 * 
	 * @param {RedisClient} redisClient 
	 */
	static _promisifyClient(redisClient)
	{
		const {promisify} = require("util");

		redisClient.getAsync = promisify(redisClient.get);
		redisClient.setAsync = promisify(redisClient.set);
		redisClient.delAsync = promisify(redisClient.del);
		redisClient.flushdbAsync = promisify(redisClient.flushdb);
	}

	/**
	 * Sets handlers for Redis client.
	 */
	static _setClientEventHandlers(redisClient)
	{
		redisClient.on("error", this._onErrorHandler);
	}

	/**
	 * Handle for error event on Redis client.
	 * 
	 * @param {Error} error 
	 */
	static _onErrorHandler(error)
	{
		console.error(`${this.name} Error:`, error);
	}

	/**
	 * Contains a list of required properties for the config object.
	 * 
	 * @type {Array}
	 */
	static get _requiredConfigProperties() 
	{ 
		let arrRequiredProperties = super._requiredConfigProperties;
		arrRequiredProperties.push("host");
		arrRequiredProperties.push("port");

		return arrRequiredProperties; 
	}
};