class ConsoleFormat
{
	constructor({
		clusterMode = false,
		verbosityLevel = ConsoleFormat.VERBOSITY_DEBUG,
		useFullTextColors = false,
		useColors
	} = {})
	{
		this.clusterMode = clusterMode;
		this.verbosityLevel = verbosityLevel;
		this.useColors = useColors || process.stdout.isTTY;
		this.useFullTextColors = useFullTextColors;

		this._bGlobalConsoleFunctionsFetched = false;
	}

	get clusterMode() { return this._bClusterMode; }
	set clusterMode(value)
	{
		this._bClusterMode = !!value;
	}

	get verbosity() { return this._strVerbosity; }
	set verbosity(value)
	{
		if(!this.allowedVerbosity.includes(value))
		{
			throw new TypeError(`Invalid verbosity ${JSON.stringify(value)}. Allowed ones are: ${JSON.stringify(this.allowedVerbosity)}.`);
		}
		this._strVerbosity = value;
	}

	get useColors() { return this._bUseColors; }
	set useColors(value)
	{
		this._bUseColors = !!value;
	}

	get useFullTextColors() { return this._bUseFullTextColors; }
	set useFullTextColors(value)
	{
		this._bUseFullTextColors = !!value;
	}

	get verbosityLevels()
	{
		return {
			[this.VERBOSITY_ERROR]: 0,
			[this.VERBOSITY_WARN]: 1,
			[this.VERBOSITY_INFO]: 2,
			[this.VERBOSITY_SUCCESS]: 2,
			[this.VERBOSITY_LOG]: 3,
			[this.VERBOSITY_DEBUG]: 10
		}
	}

	get allowedVerbosity()
	{
		if(!this._arrAllowedVerbosity)
		{
			this._arrAllowedVerbosity = Object.keys(this.verbosityLevels);
		}

		return this._arrAllowedVerbosity;
	}

	getNewInstance(objOptions)
	{
		return new ConsoleFormat(objOptions);
	}

	getConsolePrefix(strVerbosity = ConsoleFormat.VERBOSITY_DEBUG)
	{
		let strClusterRole = "";
		if(this.clusterMode)
		{
			const cluster = require("cluster");
			strClusterRole = cluster.isMaster ? "[Master]" : `[W #${cluster.worker.id}]`;
			strClusterRole = strClusterRole.padEnd(8);
		}

		let strDate = (new Date()).toISOString().replace("T", " ").replace("Z", " UTC");

		return `${strDate} ${String(`[PID ${process.pid}]`).padEnd(10)} ${strClusterRole} ${(`${strVerbosity}`).padEnd(7)}`;
	}

	consoleWrapper(strVerbosity, ...args)
	{
		if(this.verbosityLevels[strVerbosity] > this.verbosityLevels[this.verbosity])
		{
			return;
		}

		let bUseFullTextColor = false;
		let fnConsoleFunction = null;
		const bHasStringFormat = args.length > 0 && typeof args[0] === "string" && args[0].includes("%");

		let strConsoleColor = this.CONSOLE_COLOR_FG_WHITE;
		switch(strVerbosity)
		{
			case this.VERBOSITY_DEBUG:
				fnConsoleFunction = this.fnOriginalConsoleDebug;
				strConsoleColor = this.CONSOLE_COLOR_FG_WHITE;
				break;
			case this.VERBOSITY_LOG:
				fnConsoleFunction = this.fnOriginalConsoleLog;
				strConsoleColor = this.CONSOLE_COLOR_FG_CYAN;
				break;
			case this.VERBOSITY_INFO:
				fnConsoleFunction = this.fnOriginalConsoleInfo;
				strConsoleColor = this.CONSOLE_COLOR_FG_BLUE;
				break;
			case this.VERBOSITY_SUCCESS:
				fnConsoleFunction = this.fnOriginalConsoleLog;
				strConsoleColor = this.CONSOLE_COLOR_FG_GREEN;
				break;
			case this.VERBOSITY_WARN:
				fnConsoleFunction = this.fnOriginalConsoleWarn;
				strConsoleColor = this.CONSOLE_COLOR_FG_YELLOW;
				break;
			case this.VERBOSITY_ERROR:
				fnConsoleFunction = this.fnOriginalConsoleError;
				bUseFullTextColor = true;
				strConsoleColor = this.CONSOLE_COLOR_FG_RED;
				break;
			default:
				fnConsoleFunction = this.fnOriginalConsoleLog;
				break;
		}

		if(!this.useColors)
		{
			if(bHasStringFormat)
			{
				fnConsoleFunction(this.getConsolePrefix(strVerbosity));
				fnConsoleFunction(...args);
			}
			else
			{
				fnConsoleFunction(this.getConsolePrefix(strVerbosity), ...args);
			}
		}
		else if(bUseFullTextColor || this.useFullTextColors)
		{
			if(bHasStringFormat)
			{
				fnConsoleFunction(`${strConsoleColor}%s`, this.getConsolePrefix(strVerbosity));
				fnConsoleFunction(...args);
				fnConsoleFunction(this.CONSOLE_COLOR_RESET);
			}
			else
			{
				fnConsoleFunction(`${strConsoleColor}%s`, this.getConsolePrefix(strVerbosity), ...args, this.CONSOLE_COLOR_RESET);
			}
		}
		else
		{
			if(bHasStringFormat)
			{
				fnConsoleFunction(`${strConsoleColor}%s${this.CONSOLE_COLOR_RESET}`, this.getConsolePrefix(strVerbosity));
				fnConsoleFunction(...args);
			}
			else
			{
				fnConsoleFunction(`${strConsoleColor}%s${this.CONSOLE_COLOR_RESET}`, this.getConsolePrefix(strVerbosity), ...args);
			}
		}
	}

	setupGlobalConsole()
	{
		this._setupConsole(console);
	}

	resetGlobalConsole()
	{
		this.resetConsole(console);
	}

	resetConsole(consoleToReset)
	{
		this._fetchConsoleOriginalFunctions();

		consoleToReset.log = this.fnOriginalConsoleLog;
		consoleToReset.debug = this.fnOriginalConsoleDebug;
		consoleToReset.info = this.fnOriginalConsoleInfo;
		consoleToReset.warn = this.fnOriginalConsoleWarn;
		consoleToReset.error = this.fnOriginalConsoleError;
		consoleToReset.success = this.fnOriginalConsoleLog;
	}

	setupNewConsole(outStream, errStream)
	{
		const newConsole = new console.Console(outStream, errStream);
		return this._setupConsole(newConsole);
	}

	_fetchConsoleOriginalFunctions()
	{
		if(!this._bGlobalConsoleFunctionsFetched)
		{
			this.fnOriginalConsoleLog = this.fnOriginalConsoleLog || console.log;
			this.fnOriginalConsoleWarn = this.fnOriginalConsoleWarn || console.warn;
			this.fnOriginalConsoleError = this.fnOriginalConsoleError || console.error;
			this.fnOriginalConsoleDebug = this.fnOriginalConsoleDebug || console.debug;
			this.fnOriginalConsoleInfo = this.fnOriginalConsoleInfo || console.info;
			this._bGlobalConsoleFunctionsFetched = true;
		}
	}

	_setupConsole(consoleToSetup)
	{
		this._fetchConsoleOriginalFunctions();

		consoleToSetup.log = (...args) => {
			this.consoleWrapper(this.VERBOSITY_LOG, ...args);
		};
		consoleToSetup.warn = (...args) => {
			this.consoleWrapper(this.VERBOSITY_WARN, ...args);
		};
		consoleToSetup.error = (...args) => {
			this.consoleWrapper(this.VERBOSITY_ERROR, ...args);
		};

		consoleToSetup.debug = (...args) => {
			this.consoleWrapper(this.VERBOSITY_DEBUG, ...args);
		};
		consoleToSetup.info = (...args) => {
			this.consoleWrapper(this.VERBOSITY_INFO, ...args);
		};
		//Add success for prettier logs.
		consoleToSetup.success = (...args) => {
			this.consoleWrapper(this.VERBOSITY_SUCCESS, ...args);
		};

		return consoleToSetup;
	}

	//https://stackoverflow.com/questions/9781218/how-to-change-node-jss-console-font-color
	get CONSOLE_COLOR_RESET() { return "\x1b[0m"; };
	get CONSOLE_COLOR_BRIGHT() { return "\x1b[1m"; };
	get CONSOLE_COLOR_DIM() { return "\x1b[2m"; };
	get CONSOLE_COLOR_UNDERSCORE() { return "\x1b[4m"; };
	get CONSOLE_COLOR_BLINK() { return "\x1b[5m"; };
	get CONSOLE_COLOR_REVERSE() { return "\x1b[7m"; };
	get CONSOLE_COLOR_HIDDEN() { return "\x1b[8m"; };

	get CONSOLE_COLOR_FG_BLACK() { return "\x1b[30m"; };
	get CONSOLE_COLOR_FG_RED() { return "\x1b[31m"; };
	get CONSOLE_COLOR_FG_GREEN() { return "\x1b[32m"; };
	get CONSOLE_COLOR_FG_YELLOW() { return "\x1b[33m"; };
	get CONSOLE_COLOR_FG_BLUE() { return "\x1b[34m"; };
	get CONSOLE_COLOR_FG_MAGENTA() { return "\x1b[35m"; };
	get CONSOLE_COLOR_FG_CYAN() { return "\x1b[36m"; };
	get CONSOLE_COLOR_FG_WHITE() { return "\x1b[37m"; };

	get CONSOLE_COLOR_BG_BLACK() { return "\x1b[40m"; };
	get CONSOLE_COLOR_BG_RED() { return "\x1b[41m"; };
	get CONSOLE_COLOR_BG_GREEN() { return "\x1b[42m"; };
	get CONSOLE_COLOR_BG_YELLOW() { return "\x1b[43m"; };
	get CONSOLE_COLOR_BG_BLUE() { return "\x1b[44m"; };
	get CONSOLE_COLOR_BG_MAGENTA() { return "\x1b[45m"; };
	get CONSOLE_COLOR_BG_CYAN() { return "\x1b[46m"; };
	get CONSOLE_COLOR_BG_WHITE() { return "\x1b[47m"; };


	get VERBOSITY_ERROR() { return "ERROR"; }
	get VERBOSITY_WARN() { return "WARN"; }
	get VERBOSITY_INFO() { return "INFO"; }
	get VERBOSITY_SUCCESS() { return "SUCCESS"; }
	get VERBOSITY_LOG() { return "LOG"; }
	get VERBOSITY_DEBUG() { return "DEBUG"; }
};

module.exports = new ConsoleFormat();
