# ConsoleFormat

Singleton used to **alter** global `console` **functions** or any other `console.Console` instance. It adds specific **color** for each function output and adds the following information:

* UTC Date
* process PID
* process cluster role **if in cluster mode**: `Master` or `Worker` with it's ID
* verbosity level: `LOG`, `WARN`, `ERROR`, `INFO`, `DEBUG`
* adds additional function `console.success(...args)` - for extra verbosity and colors (verbosity: `SUCCESS`)

It also provides codes for manipulating terminal output color. By default, colors are only used when the `stdout` is a terminal (is a `TTY`).

## Options

Property name       | Type      | Default                           | Description
---                 | ---       | ---                               | --- 
clusterMode         | `boolean` | `false`                           | Show cluster role if enabled. `Master` or `W #{workerID}`.
verbosity           | `string`  | `ConsoleFormat.VERBOSITY_DEBUG`   | Specifies **maximum** verbosity.
useColors           | `boolean` | `true` if `stdout` is **TTY**     | Adds terminal colors to the output.
useFullTextColors   | `boolean` | `false`                           | Sets the full output with the specific color. Default, only the prefix is colored.

**Notice:** Any property can be modified after the setup was run, and the changes will affect the following `console functions` calls.

## Verbosity

Verbosity                           | Value     | Level
---                                 | ---       | ---
`ConsoleFormat.VERBOSITY_ERROR`     | `ERROR`   | 0
`ConsoleFormat.VERBOSITY_WARN`      | `WARN`    | 1
`ConsoleFormat.VERBOSITY_INFO`      | `INFO`    | 2
`ConsoleFormat.VERBOSITY_SUCCESS`   | `SUCCESS` | 2
`ConsoleFormat.VERBOSITY_LOG`       | `LOG`     | 3
`ConsoleFormat.VERBOSITY_DEBUG`     | `DEBUG`   | 10

## Colors

Modifiers:

* `CONSOLE_COLOR_RESET`
* `CONSOLE_COLOR_BRIGHT`
* `CONSOLE_COLOR_DIM`
* `CONSOLE_COLOR_UNDERSCORE`
* `CONSOLE_COLOR_BLINK`
* `CONSOLE_COLOR_REVERSE`
* `CONSOLE_COLOR_HIDDEN`

Foreground:

* `CONSOLE_COLOR_FG_BLACK`
* `CONSOLE_COLOR_FG_RED`
* `CONSOLE_COLOR_FG_GREEN`
* `CONSOLE_COLOR_FG_YELLOW`
* `CONSOLE_COLOR_FG_BLUE`
* `CONSOLE_COLOR_FG_MAGENTA`
* `CONSOLE_COLOR_FG_CYAN`
* `CONSOLE_COLOR_FG_WHITE`

Background:

* `CONSOLE_COLOR_BG_BLACK`
* `CONSOLE_COLOR_BG_RED`
* `CONSOLE_COLOR_BG_GREEN`
* `CONSOLE_COLOR_BG_YELLOW`
* `CONSOLE_COLOR_BG_BLUE`
* `CONSOLE_COLOR_BG_MAGENTA`
* `CONSOLE_COLOR_BG_CYAN`
* `CONSOLE_COLOR_BG_WHITE`

## Example

### Global console

```js
const { ConsoleFormat } = require("jskippy-node");

//Use helper color codes directly
console.log(ConsoleFormat.CONSOLE_COLOR_BG_BLUE, ConsoleFormat.CONSOLE_COLOR_FG_WHITE, "Something here", ConsoleFormat.CONSOLE_COLOR_RESET);

//Setup global console
ConsoleFormat.setupGlobalConsole();
console.log("test");
console.warn("test");
console.error("test");
console.info("test");
console.debug("test");
console.success("test");

//Reset global console
ConsoleFormat.resetGlobalConsole();
console.log("test");
console.warn("test");
console.error("test");
console.info("test");
console.debug("test");
console.success("test"); //This will remain as a alias to console.log, so that the code will not throw an error.

```

### Custom console

```js
//Optionally, a new instance of ConsoleFormat can be created, so that the global console will not be affected by modifying properties on the custom console.
const newConsoleFormat = ConsoleFormat.getNewInstance(/* objOptions */);

//New Console setup
const newConsole = newConsoleFormat.setupNewConsole(process.stdout, process.stderr); //Or any other WriteStream
newConsole.log("test");
newConsole.warn("test");
newConsole.error("test");
newConsole.info("test");
newConsole.debug("test");
newConsole.success("test");

ConsoleFormat.resetConsole(newConsole);
newConsole.log("test");
newConsole.warn("test");
newConsole.error("test");
newConsole.info("test");
newConsole.debug("test");
newConsole.success("test");

//New Console with output on log files
const output = fs.createWriteStream('./stdout.log');
const errorOutput = fs.createWriteStream('./stderr.log');

const logger = newConsoleFormat.setupNewConsole(process.stdout, process.stderr); //Or any other WriteStream
logger.log("test");
logger.warn("test");
logger.error("test");
logger.info("test");
logger.debug("test");
logger.success("test");
// use it like console
const count = 5;
logger.log("count: %d", count);
```