const { Validators, Cache, ConsoleFormat } = require("../index");

async function test_run_all()
{
	try
	{
		await test_cache();
		test_validators();
		test_console_format();
		process.exit(0);
	}
	catch(error)
	{
		console.error(error);
		process.exit(1);
	}
}

async function test_cache()
{
	Cache.CacheSingleton.config = {
		engine: "redis",
		redis: {
			//https://www.npmjs.com/package/redis#rediscreateclient
			host: "192.168.137.2",
			port: 6379,
			sandboxCacheKey: "SOME_CACHE_KEY"
		}
	};

	test_cache_run_for("ceva", {ceva: 1}, 1, 2);
	test_cache_run_for("ceva1", true, 60);
	test_cache_run_for("ceva2", 123, 60);
	test_cache_run_for("ceva3", "altceva", 60);
	test_cache_run_for("ceva4", 12.4, 60);
}

async function test_cache_run_for(strKey, mxValue, nExpireSeconds, nFetchAfterSeconds = 1, bForceExpire = false)
{
	//TODO Log OK or throw error
	try
	{
		function cache()
		{
			return Cache.CacheSingleton.getInstance();
		}

		async function fetch(strKey)
		{
			let result = await cache().fetch(strKey);

			if(result.hasValue())
			{
				console.log("CACHE", result.value);
			}
			else
			{
				console.log("Cache MISS");
			}
		}

		await cache().store(strKey, mxValue, nExpireSeconds);

		console.log("Cache test start...");

		setTimeout(
			async () => {
				await fetch(strKey);
			},
			nFetchAfterSeconds * 1000
		);

		if(bForceExpire)
		{
			setTimeout(
				async () => {
					await cache().expire(strKey);
					await fetch(strKey);
				},
				1000
			);
		}
	}
	catch(error)
	{
		console.error(error);
	}
}

function test_validators()
{
	console.log("--------- DomainNameValidator -----------");
	Validators.DomainName.DomainNameValidator.test();
	console.log("--------- EmailAddressValidator -----------");
	Validators.EmailAddress.EmailAddressValidator.test();
	console.log("--------- ISCSINameValidator -----------");
	Validators.ISCSIName.ISCSINameValidator.test();
	console.log("--------- TimestampValidator -----------");
	Validators.Timestamp.TimestampValidator.test();
	console.log("--------- URLValidator -----------");
	Validators.URL.URLValidator.test();
	console.log("--------- ProductSubdomainName -----------");
	Validators.ProductSubdomainName.ProductSubdomainNameValidator.test();
}

function test_console_format()
{
	console.log(ConsoleFormat.CONSOLE_COLOR_BG_BLUE, ConsoleFormat.CONSOLE_COLOR_FG_WHITE, "BEGIN test_console_format", ConsoleFormat.CONSOLE_COLOR_RESET);
	console.log("---------- SETUP GLOBAL CONSOLE -------------");
	ConsoleFormat.setupGlobalConsole();
	_runConsoleOutputTests();
	
	console.log("---------- VERBOSITY_WARN -------------");
	ConsoleFormat.verbosity = ConsoleFormat.VERBOSITY_WARN;
	_runConsoleOutputTests();
	ConsoleFormat.verbosity = ConsoleFormat.VERBOSITY_DEBUG;
	
	console.log("---------- USE COLORS -------------");
	ConsoleFormat.useColors = true;
	_runConsoleOutputTests();

	console.log("---------- CLUSTER TRUE -------------");
	ConsoleFormat.clusterMode = true;
	_runConsoleOutputTests();

	console.log("---------- CLUSTER FALSE -------------");
	ConsoleFormat.clusterMode = false;
	_runConsoleOutputTests();
	
	console.log("---------- USE FULL TEXT COLORS -------------");
	ConsoleFormat.useFullTextColors = true;
	_runConsoleOutputTests();

	console.log("---------- USE FULL TEXT COLORS FALSE -------------");
	ConsoleFormat.useFullTextColors = false;
	_runConsoleOutputTests();

	console.log("---------- USE COLORS FALSE -------------");
	ConsoleFormat.useColors = false;
	_runConsoleOutputTests();
	
	console.log("---------- RESET GLOBAL CONSOLE -------------");
	ConsoleFormat.resetGlobalConsole();
	_runConsoleOutputTests();
	
	//New Console
	console.log("---------- SETUP NEW CONSOLE -------------");
	const newConsoleFormat = ConsoleFormat.getNewInstance();
	const newConsole = newConsoleFormat.setupNewConsole(process.stdout, process.stderr);
	_runConsoleOutputTests(newConsole);

	console.log("---------- RESET NEW CONSOLE -------------");
	ConsoleFormat.resetConsole(newConsole);
	_runConsoleOutputTests(newConsole);

	console.log(ConsoleFormat.CONSOLE_COLOR_BG_GREEN, ConsoleFormat.CONSOLE_COLOR_FG_BLACK, "[OK] test_console_format", ConsoleFormat.CONSOLE_COLOR_RESET);
}

function _runConsoleOutputTests(aConsole = null)
{
	aConsole = aConsole || console;
	aConsole.log("test");
	aConsole.warn("test");
	aConsole.error("test");
	aConsole.info("test");
	aConsole.debug("test");
	aConsole.success("test");

	aConsole.log("%s %d", "ceva", 123);
}

test_run_all();
