const sleep = require("sleep-promise");
const BDKAgentClient = require("../EndpointClients/BDKAgent");
const AFCMakeCallEndpointBase = require("./AFCMakeCallEndpointBase");
const powerSingleton = require("../Modules").PowerSingleton;

const ServerInfoGatherer = require("../../BDKAgentModules/ServerInfoGatherer").ServerInfoGathererSingleton;
const ServerCleanup = require("../../BDKAgentModules/ServerCleanup");
const ServerRegistering = require("../../BDKAgentModules/ServerRegistering");
const {Encryption} = require("encryption");

module.exports =
	class BDKAgent extends AFCMakeCallEndpointBase
	{
		constructor(agentIdentity)
		{
			super(
				/*strName*/ "BDKAgent",
				/*strPath*/ "/api-ws/BDKAgent",
				/*objReflection*/ {},
				/*classReverseCallsClient*/ BDKAgentClient
			);

			this._agentIdentity = agentIdentity;
			this._serverEncryption = new Encryption(this._agentIdentity.serverEncryptionRotatingKeys);
			this._objAFCPromises = {};
			this._bUpdatePrerequisiteFinished = false;
		}

		/**
		 * @param {JSONRPC.IncomingRequest} incomingRequest
		 * @param {boolean} bForce = false
		 * @param {number} nTimeoutMilliseconds = 0
		 * @returns {string}
		 */
		async bdk_agent_reboot(incomingRequest, bForce, nTimeoutMilliseconds)
		{
			await powerSingleton.powerOff(/*bReboot*/ true, bForce, nTimeoutMilliseconds);

			return "Server will now reboot.";
		}

		/**
		 * @param {JSONRPC.IncomingRequest} incomingRequest
		 * @param {boolean} bForce = false
		 * @param {number} nTimeoutMilliseconds = 0
		 * @returns {string}
		 */
		async bdk_agent_shutdown(incomingRequest, bForce, nTimeoutMilliseconds)
		{
			await powerSingleton.powerOff(/*bReboot*/ false, bForce, nTimeoutMilliseconds);

			return "Server will now shutdown.";
		}

		/**
		 * @param {JSONRPC.IncomingRequest} incomingRequest
		 * @param {number} nSleepSeconds
		 * @returns {string}
		 */
		async bdk_agent_test(incomingRequest, nSleepSeconds)
		{
			const strStartTime = new Date().toISOString();
			await sleep(nSleepSeconds * 1000);

			return "Hello! Test start time: " + strStartTime + ". End time: " + (new Date().toISOString());
		}


		/**
		 * Collects various hardware information needed to update the newly created server in the database
		 * 
		 * @param {JSONRPC.IncomingRequest} incomingRequest
		 * @returns {Object}
		 */
		async bdk_agent_server_info_collect(incomingRequest)
		{
			return await ServerInfoGatherer.gatherAllInfo();
		}


		/**
		 * Execute the required cleanup stages (example: wipe the local disks of the server)
		 * 
		 * @param {JSONRPC.IncomingRequest} incomingRequest
		 * @param {Object} objServer
		 * @param {Array<Object>} arrServerInterfaces
		 * @param {Array} objIPOOB
		 * @param {Array} arrDNSServers
		 * @param {Array} arrNTPServers
		 * @returns {string}
		 */
		async bdk_agent_server_cleanup(incomingRequest, objServer, arrServerInterfaces, objIPOOB, arrDNSServers, arrNTPServers)
		{
			return await ServerCleanup.BDKCleanup.run(objServer, arrServerInterfaces, objIPOOB, arrDNSServers, arrNTPServers);
		}

		/**
	 	 * This function will set up IPMI for the server.
		 * Credentials and IP addresses should be created or allocated before calling this function.
		 * 
		 * @param {JSONRPC.IncomingRequest} incomingRequest
		 * @param {Object} [objConfig]
		 * @returns {Promise<number>}
		 */
		async bdk_agent_IPMI_setup(incomingRequest, objConfig)
		{
			return await ServerRegistering.IPMISetup.run(objConfig, this._serverEncryption);
		}
		

		/**
		 * Exits the current process after 2 seconds using setTimeout().
		 * Should return before the setTimeout's callback executes. Two seconds should be enough time to return.
		 *
		 * @param {JSONRPC.IncomingRequest} incomingRequest
		 * 
		 * @returns {string} 
		 */
		async bdk_agent_process_exit(incomingRequest)
		{
			setTimeout(
				() =>
				{
					try
					{
						console.log("Will process.exit(1) now.");
						process.exit(1);
					}
					catch (error)
					{
						console.error(error);
					}
				},
				2000
			);
			
			return "Will process.exit(1) in 2 seconds";
		}
	};
