// @ts-check

const os = require("os");
const si = require("systeminformation");

const UtilityOSInfo = require("./UtilityOSInfo");


class ServerInfoGatherer
{
	constructor()
	{
		this._objUnprocessedServerInfo = null;
		
		this._objGeneralInfo = null;
		this._objDiskInfo = null;
		this._objNetworkInterfaceInfo = null;
		this._objBiosInfo = null;
		this._objCPUInfo = null;
		this._objMemoryInfo = null;

		this._strOSPlatform = os.platform();
	}

	/**
	 * Returns an object with server information.
	 * 
	 * @returns {Promise<Object>}
	 */
	async gatherAllInfo()
	{
		if(this._objUnprocessedServerInfo === null)
		{
			await this._generateUnprocessedServerInfo();
		}

		this._objAllServerInfo = {
			...this._objGeneralInfo ? this._objGeneralInfo : await this.gatherGeneralInfo(),
			...this._objBiosInfo ? this._objBiosInfo : await this.gatherBiosInfo(),
			...this._objCPUInfo = this._objCPUInfo ? this._objCPUInfo : await this.gatherCPUInfo(),
			...this._objMemoryInfo ? this._objMemoryInfo : await this.gatherMemoryInfo(),
			...this._objDiskInfo ? this._objDiskInfo : await this.gatherDiskInfo(),
			...this._objNetworkInterfaceInfo ? this._objNetworkInterfaceInfo : await this.gatherNetworkInterfaceInfo()
		};
			
		return this._objAllServerInfo;
	}

	/**
	 * Returns an object with main informations about the server.
	 * Example: UUID, ServerName, etc.
	 * 
	 * @returns {Promise<Object>}
	 */
	async gatherGeneralInfo()
	{
		if(this._objUnprocessedServerInfo === null)
		{
			await this._generateUnprocessedServerInfo();
		}

		if(this._objGeneralInfo !== null)
		{
			return this._objGeneralInfo;
		}

		this._objGeneralInfo = {};

		if(this._strOSPlatform === "linux")
		{
			try
			{
				this._objGeneralInfo.server_uuid = this._getSpecificInfoFromPath(["Lshw", "configuration", "uuid"]);
				this._objGeneralInfo.server_product_name = this._getSpecificInfoFromPath(["Lshw", "product"]);
				this._objGeneralInfo.server_vendor = this._getSpecificInfoFromPath(["Lshw", "vendor"]);
				this._objGeneralInfo.server_serial_number = this._getSpecificInfoFromPath(["Lshw", "serial"]);
				this._objGeneralInfo.slot = this._getSpecificInfoFromPath(["Lshw", "core", "slot"]);
				// this._objGeneralInfo.VendorSKUID = this._getSpecificInfoFromPath(["Lshw", ]); // Not in lshw info
				this._objGeneralInfo.server_ipmi_channel = null;
			}
			catch(error)
			{
				console.error(error);
			}
		}

		if(this._strOSPlatform === "win32")
		{
			// TODO Implement
		}

		return this._objGeneralInfo;
	}

	/**
	 * Returns an object with informations regarding the bios
	 * 
	 * @returns {Promise<Object>}
	 */
	async gatherBiosInfo()
	{
		if(this._objUnprocessedServerInfo === null)
		{
			await this._generateUnprocessedServerInfo();
		}

		if(this._objBiosInfo !== null)
		{
			return this._objBiosInfo;
		}
		
		this._objBiosInfo = {};
		
		if(this._strOSPlatform === "linux")
		{
			try
			{
				this._objBiosInfo.server_bios_vendor = this._getSpecificInfoFromPath(["Lshw", "children", 0, "children", 0, "vendor"]);
				this._objBiosInfo.server_bios_version = this._getSpecificInfoFromPath(["Lshw", "children", 0, "children", 0, "version"]);
				this._objBiosInfo.enabled_efi_variables = this._getSpecificInfoFromPath(["EFIbootmgr", "enabled_efi_variables"]);
			}
			catch(error)
			{
				console.error(error);
			}
		}

		if(this._strOSPlatform === "win32")
		{
			// TODO Implement
		}

		return this._objBiosInfo;
	}

	/**
	 * Returns an object with information regarding the CPU
	 * 
	 * @returns {Promise<Object>}
	 */
	async gatherCPUInfo()
	{
		if(this._objUnprocessedServerInfo === null)
		{
			await this._generateUnprocessedServerInfo();
		}

		if(this._objCPUInfo !== null)
		{
			return this._objCPUInfo;
		}
		
		this._objCPUInfo = {};

		if(this._strOSPlatform === "linux")
		{
			const objProcessorCpuMarkMap = {
				"Intel(R) Xeon(R) CPU E31240 @ 3.30GHz": "8324",
				"Intel(R) Xeon(R) CPU E5-2430 0 @ 2.20GHz": "13624",
				"Intel(R) Xeon(R) CPU           E5410  @ 2.33GHz": "3203",
				"Intel(R) Xeon(R) CPU E3-1230 V2 @ 3.30GHz": "8887",
				"Intel(R) Xeon(R) CPU           L5410  @ 2.33GHz": "3921",
				"Intel(R) Xeon(R) CPU            5130  @ 2.00GHz": "1404"
			};

			try
			{
				this._objCPUInfo.server_processor_count = this._getSpecificInfoFromPath(["Lshw", "children", 0, "children"]).reduce((acc, value) => acc + +(value.description === "CPU"), 0);
				
				// For VMs lshw could display the wrong number of CPUs
				if(this._objCPUInfo.server_processor_count >= 128)
				{
					this._objCPUInfo.server_processor_count = this._getSpecificInfoFromPath(["Lscpu", "server_processor_count"]);
				}

				this._objCPUInfo.server_processor_core_count = this._getSpecificInfoFromPath(["Lshw", "children", 0, "children", 1, "configuration", "cores"]);
				this._objCPUInfo.server_processor_enabled_core_count = this._getSpecificInfoFromPath(["Lshw", "children", 0, "children", 1, "configuration", "enabledcores"]);
				this._objCPUInfo.server_processor_threads = this._getSpecificInfoFromPath(["Lscpu", "server_processor_threads"]);
				this._objCPUInfo.server_processor_name = this._getSpecificInfoFromPath(["Lshw", "children", 0, "children", 1, "version"]);
				this._objCPUInfo.server_processor_core_mhz = this._getSpecificInfoFromPath(["Lshw", "children", 0, "children", 1, "size"]) / 1000000;
				this._objCPUInfo.server_processor_cpu_mark = objProcessorCpuMarkMap[this._objCPUInfo.server_processor_name] || null;
			}
			catch(error)
			{
				console.error(error.stack);
			}

		}

		if(this._strOSPlatform === "win32")
		{
			// TODO Implement
		}

		return this._objCPUInfo;
	}

	/**
	 * Returns an object with the information regarding the main memory
	 * 
	 * @returns {Promise<Object>}
	 */
	async gatherMemoryInfo()
	{
		if(this._objUnprocessedServerInfo === null)
		{
			await this._generateUnprocessedServerInfo();
		}

		if(this._objMemoryInfo !== null)
		{
			return this._objMemoryInfo;
		}

		this._objMemoryInfo = {};

		if(this._strOSPlatform === "linux")
		{
			try
			{
				let nRAMMemorySizeGB = parseInt(this._getSpecificInfoFromPath(["Lshw", "children", 0, "children", 2, "size"]), 10);
		
				// If the size is 0 count the size of every memory bank
				if(nRAMMemorySizeGB === 0)
				{
					const arrMemoryBanks = this._getSpecificInfoFromPath(["Lshw", "children", 0, "children", 2, "children"]);
					nRAMMemorySizeGB = arrMemoryBanks.reduce((acc, value) => acc + value.size ? parseInt(value.size, 10) : 0);
				}
				
				this._objMemoryInfo.server_ram_gbytes = Math.round(nRAMMemorySizeGB / (1024 * 1024 * 1024));
			}
			catch(error)
			{
				console.error(error);
			}
		}

		if(this._strOSPlatform === "win32")
		{
			// TODO Implement
		}


		return this._objMemoryInfo;
	}

	/**
	 * Returns an object with informations regarding the disks
	 * As there could be more than one disk, the specific information objects are grouped in an array.
	 * Properties from a disk object:
	 * [
	 * 		server_disk_size_gb,
	 * 		server_disk_type,
	 * 		server_disk_serial,
	 * 		server_disk_model,
	 * 		server_disk_vendor
	 * ]
	 * 
	 * @returns {Promise<Object>}
	 */
	async gatherDiskInfo()
	{
		if(this._objUnprocessedServerInfo === null)
		{
			await this._generateUnprocessedServerInfo();
		}

		if(this._objDiskInfo !== null)
		{
			return this._objDiskInfo;
		}

		this._objDiskInfo = {};

		if(this._strOSPlatform === "linux")
		{
			try
			{
				// Because lshw doesn't have NVMe disk detection, other tools are further required
				this._objDiskInfo.diskCollection = [
					...(this._getSpecificInfoFromPath(["DiskDetection", "disks"]) || []),
					...(this._getSpecificInfoFromPath(["NVMe", "disks"]) || [])
				];

				if(!this._objDiskInfo.diskCollection.length)
				{
					this._objDiskInfo.server_disk_count = 0;
				}
				else
				{
					this._objDiskInfo.server_disk_count = this._objDiskInfo.diskCollection.length;

					// IMPORTANT RESTRAINT: Using the smallest size for global size
					this._objDiskInfo.server_disk_size_gb = this._objDiskInfo.diskCollection.reduce((nMinDiskSize, objDisk) => 
					{
						return parseInt(objDisk.server_disk_size_gb, 10) > nMinDiskSize 
							? parseInt(objDisk.server_disk_size_gb, 10) 
							: nMinDiskSize;
					}, parseInt(this._objDiskInfo.diskCollection[0].server_disk_size_gb, 10));
	
					this._objDiskInfo.server_disk_type = this._objDiskInfo.diskCollection[0].server_disk_type;
				}
			}
			catch(error)
			{
				console.error(error);
			}
		}

		if(this._strOSPlatform === "win32")
		{
			// TODO Implement
		}

		return this._objDiskInfo;
	}

	/**
	 * Returns an object with informations regarding the Network Interfaces
	 * 
	 * @returns {Promise<Object>} 
	 */
	async gatherNetworkInterfaceInfo()
	{
		if(this._objUnprocessedServerInfo === null)
		{
			await this._generateUnprocessedServerInfo();
		}

		if(this._objNetworkInterfaceInfo !== null)
		{
			return this._objNetworkInterfaceInfo;
		}

		this._objNetworkInterfaceInfo = {};
		const arrPCINetworkInterfaceChildren = [];

		if(this._strOSPlatform === "linux")
		{
			try
			{
				this._objNetworkInterfaceInfo.server_network_total_capacity_mbps = this._getSpecificInfoFromPath(["NetworkInterfaces", "server_network_total_capacity_mbps"]);

				try
				{
					const arrPCIChildren = this._getSpecificInfoFromPath(["Lshw", "children", 0, "children"]).filter(value => value.id === "pci")[0].children;
					
					if(Array.isArray(arrPCIChildren))
					{
						arrPCIChildren.forEach(objPCISlots => 
						{
							if(objPCISlots.id.split(":")[0] === "pci" && objPCISlots.children)
							{
								const arrPCIInnerChildren = objPCISlots.children;
	
								arrPCIInnerChildren.forEach(objPCISlot => 
								{
									if(objPCISlot.id === "network")
									{
										arrPCINetworkInterfaceChildren.push(objPCISlot);
									}
								});
							}
						});
					}
					else
					{
						throw new Error("Selected property from lshw PCI section is not an array.");
					}
				}
				catch(error)
				{
					console.error(error);
				}

				this._objNetworkInterfaceInfo.NICDetails = {};
				let nLSHWServerNetworkTotalCapacity = 0;

				// Details about each NIC are grouped by MAC address
				arrPCINetworkInterfaceChildren.forEach(objNetworkInterface => 
				{
					const strNetworkMAC = objNetworkInterface.serial;
					nLSHWServerNetworkTotalCapacity += objNetworkInterface.size ? objNetworkInterface.size / 1000000 : objNetworkInterface.capacity / 100000;
	
					this._objNetworkInterfaceInfo.NICDetails[strNetworkMAC] = {};
					this._objNetworkInterfaceInfo.NICDetails[strNetworkMAC]["vendor"] = objNetworkInterface.vendor || "Unspecified vendor";
					this._objNetworkInterfaceInfo.NICDetails[strNetworkMAC]["product"] = objNetworkInterface.product || "Unspecified product";
					this._objNetworkInterfaceInfo.NICDetails[strNetworkMAC]["logicalname"] = objNetworkInterface.logicalname || "Unspecified logicalname";
				});
	
				if(!this._objNetworkInterfaceInfo["server_network_total_capacity_mbps"])
				{
					this._objNetworkInterfaceInfo["server_network_total_capacity_mbps"] = nLSHWServerNetworkTotalCapacity;
				}

				this._objNetworkInterfaceInfo.server_network_total_capacity_mbps = parseInt(this._objNetworkInterfaceInfo.server_network_total_capacity_mbps, 10);
			}
			catch(error)
			{
				console.error(error);
			}
		}

		if(this._strOSPlatform === "win32")
		{
			// TODO Implement
		}

		return this._objNetworkInterfaceInfo;
	}

	async _generateUnprocessedServerInfo()
	{
		switch (this._strOSPlatform) 
		{
			case "linux":
				this._objUnprocessedServerInfo = await UtilityOSInfo.runCommandBatch();
				break;
			case "win32":
				this._objUnprocessedServerInfo = await si.getAllData();
				break;
		}
	}


	/**
	 * Searches in the given Object for the property that is found in the path specified in arrPath.
	 * arrPath Example: ["propertyOne", 0, "propertyTwo"].
	 * If the path is wrong null is returned.
	 * 
	 * @param {Array} arrPath
	 * @param {Object|Array} objSearchable
	 * @returns {(string|Object|null)}
	 */
	_getSpecificInfoFromPath(arrPath, objSearchable = this._objUnprocessedServerInfo)
	{
		return arrPath.reduce(
			(objAccumulator, mxCurentValue) => (objAccumulator && objAccumulator[mxCurentValue]) ? objAccumulator[mxCurentValue] : null,
			objSearchable
		);
	}

	get objUnprocessedServerInfo()
	{
		return this._objUnprocessedServerInfo;
	}
}

module.exports = ServerInfoGatherer;
